# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

"""
The sitecustomize.py file is automatically loaded by Python during interpreter
initialization. Our runner script ensures that it is on the PYTHONPATH which is
sufficient to make sure it is loaded.

See https://docs.python.org/3/library/site.html for additional details
"""

### This is the very first line of code we control in the app's python process ###

# We want to minimize the amount of code we run before we register the rewriter,
# because modules imported during that time won't be fully rewritten or patched.
# We attempt to repatch imported modules, but there still could be gaps.

import sys
import os
import contrast_rewriter

MIN_SUPPORTED_VERSION = (3, 9)
MAX_SUPPORTED_VERSION = (3, 14) if os.environ.get("CONTRAST_ALLOW_PY314") else (3, 13)
# We need to check the version_info because certain installation tools bypass
# pip and setuptools, so the version check in setup.py is skipped. For example,
# the agent operator and the universal agent bundle the agent distribution and
# prepend them to PYTHONPATH directly.


def _log(msg: str) -> None:
    contrast_rewriter.log_stderr(msg, logger_name="contrast-loader")


if MIN_SUPPORTED_VERSION <= sys.version_info[:2] <= MAX_SUPPORTED_VERSION:
    _log("Initializing Contrast Python Agent")
    _log(
        f"Installation tool: {os.environ.get('CONTRAST_INSTALLATION_TOOL', 'unknown')}"
    )
    _log(f"Command: {sys.argv}")
    _log(f"PYTHONPATH: {os.environ.get('PYTHONPATH', '<not set>')}")
    _log("Registering rewriter")

    # NOTE: This must be applied prior to importing the agent itself. Do not import any
    # other modules before registering the rewriter.
    contrast_rewriter.register()

    _log("Completed rewriter registration")

    from contrast_vendor import structlog as logging
    import contrast

    logger = logging.getLogger("contrast")
    logger.info(f"Python Agent version: {contrast.__version__}")

    from contrast import loader  # noqa: E402
    from contrast.loader import instrumentation  # noqa: E402

    logger.info("Applying early instrumentation")
    instrumentation.apply_early_instrumentation()

    loader.SITECUSTOMIZE_LOADED = True
else:
    _log(
        f"Detected unsupported environment (Python {sys.version}) - will not instrument"
    )

### Attempt to import the next sitecustomize.py file ###

loader_dir = os.path.dirname(__file__)
if loader_dir not in sys.path:
    # loader_dir will always be in sys.path when this sitecustomize is
    # being called by the Python interpreter site startup machinery.
    # If it's missing, we're being imported by some other module at
    # runtime, after the site machinery has already run. In that case,
    # don't attempt to import the next sitecustomize module because
    # that's not what an uninstrumented Python process would do.
    pass
else:
    # Unhook this sitecustomize from the import machinery.
    loader_syspath_index = sys.path.index(loader_dir)
    del sys.path[loader_syspath_index]
    this_sitecustomize = sys.modules["sitecustomize"]
    del sys.modules["sitecustomize"]

    # Attempt to import the next sitecustomize module.
    try:
        import sitecustomize  # noqa: F401
    except ImportError:
        # Only replace sitecustomize if the import fails. Otherwise,
        # an instrumented runner might assume its sitecustomize is in
        # place and raise errors if our sitecustomize were imported
        # instead.
        sys.modules["sitecustomize"] = this_sitecustomize
        sys.path.insert(loader_syspath_index, loader_dir)

    # Only catch the ImportError so that we don't swallow exceptions raised
    # by a buggy third-party sitecustomize implementation.



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast.utils.configuration_utils import (
    str_to_bool,
)


class Observe(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="observe.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from collections.abc import Mapping
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.agent.assess.utils import is_trackable


class PyMongoAction(DefaultAction):
    """
    The DefaultAction would report findings for NoSQLi if any value in the
    query were an untrusted string. This would be a false positive (and if
    it were a true finding, then the database could never hold user data).

    This action will only report findings if a dictionary exists that contains
    all untrusted keys and values. In that case, the dictionary is fully under
    user control and a malicious user could use object expansion to perform NoSQL
    injection.
    """

    unwind_source = False
    unwind_source_recurse = False

    def is_violated(self, source, required_tags, disallowed_tags, **kwargs):
        return self._is_violated(source, required_tags, disallowed_tags, **kwargs)

    def _is_violated(self, doc_field, required_tags, disallowed_tags, **kwargs):
        if isinstance(doc_field, Mapping):
            # if a JSON object contains another JSON object, we need to check
            # whether any inner JSON object is fully under user control
            object_fields = [
                field
                for field in doc_field.values()
                if isinstance(field, (Mapping, list))
            ]
            if object_fields:
                return any(
                    self._is_violated(obj, required_tags, disallowed_tags, **kwargs)
                    for obj in object_fields
                )

            # otherwise, check if this JSON object is fully under user control
            else:
                return all(
                    self._is_violated(key, required_tags, disallowed_tags)
                    and self._is_violated(value, required_tags, disallowed_tags)
                    for key, value in doc_field.items()
                    if key != "_id"  # ignore MongoDB's internal _id field
                )

        if isinstance(doc_field, list):
            return any(
                self._is_violated(doc, required_tags, disallowed_tags, **kwargs)
                for doc in doc_field
            )

        return not is_trackable(doc_field) or super().is_violated(
            doc_field, required_tags, disallowed_tags, **kwargs
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


from collections import deque
import contextlib
from dataclasses import replace
from enum import Enum
from functools import partial
from typing import Final
from collections.abc import Generator

import contrast
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.request import Request
from contrast.configuration.agent_config import AgentConfig
from contrast.configuration.config_option import DEFAULT_VALUE_SRC
from contrast.reporting.reporting_client import ReportingClient
from contrast import get_canonical_version
import contrast_fireball

from contrast.utils.configuration_utils import DEFAULT_PATHS
from contrast_vendor import structlog as logging, wrapt

logger = logging.getLogger("contrast")


def _handle_errors(return_value=None) -> wrapt.FunctionWrapper:
    """
    A decorator that catches and logs errors that occur while reporting to Contrast.

    Disabling the agent in response to authentication errors or archived applications
    is handled here.

    Errors that indicate a bug in the agent or Fireball are reported to telemetry.

    This decorator should only be used on Client methods, since it expects the
    wrapped function to be a method with AgentConfig stored on the instance.
    """

    @wrapt.function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        try:
            return wrapped(*args, **kwargs)
        except contrast_fireball.Error as e:
            if isinstance(
                e,
                (
                    contrast_fireball.ConfigurationError,
                    contrast_fireball.AuthenticationError,
                    contrast_fireball.AppArchivedError,
                ),
            ):
                # These error messages are user-facing. Log them directly without
                # the stack trace to reduce the noise in the message.
                logger.error(e.message)
            else:
                logger.error(
                    "An error occurred while reporting to Contrast", exc_info=e
                )

            if (
                isinstance(
                    e,
                    (
                        contrast_fireball.Panic,
                        contrast_fireball.ArgumentValidationError,
                        contrast_fireball.UnexpectedError,
                    ),
                )
                and contrast.TELEMETRY is not None
            ):
                contrast.TELEMETRY.report_error(e, wrapped)

            if isinstance(
                e,
                (
                    contrast_fireball.AppArchivedError,
                    contrast_fireball.AuthenticationError,
                ),
            ):
                DisableReaction.run(instance.config)

            return return_value

    return wrapper


@wrapt.function_wrapper
def _queue_if_app_uninitialized(wrapped, instance, args, kwargs):
    """
    A decorator that queues the wrapped function call if the application hasn't been
    initialized.

    This decorator should only be used on Client methods.
    """
    if not hasattr(instance, "app_id"):
        instance.queued_postinit_actions.append(partial(wrapped, *args, **kwargs))
        return None
    return wrapped(*args, **kwargs)


class Client(ReportingClient):
    """
    A client for reporting to the Contrast UI using the Fireball library.
    Fireball docs: https://fireball.prod.dotnet.contsec.com/fireball/index.html

    The client will fallback to directly reporting for endpoints that do not
    have Python bindings yet.
    """

    def __init__(self):
        self.config = None
        info = contrast_fireball.get_info()
        super().__init__(instance_id=info["reporting_instance_id"])
        self.queued_postinit_actions = deque(maxlen=10)

    @_handle_errors(return_value=False)
    def initialize_application(self, config: AgentConfig, server_type="") -> bool:
        """
        Initialize an application in the Contrast UI.

        This function must be called before any other reporting functions.
        """

        # Store config on the client for disable reaction on AppArchivedError
        self.config = config

        result = contrast_fireball.initialize_application(
            contrast_fireball.InitOptions(
                app_name=config["application.name"],
                app_path=config["application.path"],
                agent_language=contrast_fireball.AgentLanguage.PYTHON,
                agent_version=get_canonical_version(),
                server_host_name=config["server.name"],
                server_path=config["server.path"],
                server_type=server_type,
                config_paths=list(DEFAULT_PATHS),
                overrides=agent_config_to_plain_dict(config),
            )
        )
        self.app_id = result.data["app_id"]
        config.session_id = result.data["resolved_config"].get(
            "application.session_id", ""
        )
        while self.queued_postinit_actions:
            retry_action = self.queued_postinit_actions.popleft()
            retry_action()

        # This is a workaround to process the startup settings from the UI.
        # Long-term, we'll read the settings from the result, but for now
        # we want to use the well-tested direct response processing behavior.
        return super().initialize_application(config)

    @_queue_if_app_uninitialized
    @_handle_errors()
    def new_discovered_routes(self, routes: set[contrast_fireball.DiscoveredRoute]):
        """
        Report discovered routes to the Contrast UI.

        If an exception occurs, no routes are reported.
        """

        contrast_fireball.new_discovered_routes(self.app_id, list(routes))

    @_handle_errors()
    def new_observed_route(self, route: contrast_fireball.ObservedRoute):
        """
        Record an observed route.

        Routes are reported periodically in batches. This endpoint can be called multiple
        times for the same route, but Fireball will only report duplicate routes at a rate
        of once per minute to avoid overloading TeamServer.
        """

        contrast_fireball.new_observed_route(self.app_id, route)

    @_queue_if_app_uninitialized
    def new_findings(
        self,
        findings: list[contrast_fireball.AssessFinding],
        request: Request | None,
    ):
        """
        Record Assess findings.

        Findings are reported periodically in batches. Failures are handled for each
        individual finding, so that a failure in one finding does not prevent others
        from being reported.
        """
        fireball_request = request.to_fireball_assess_request() if request else None
        for finding in findings:
            self._new_finding(finding, fireball_request)

    @_handle_errors()
    def _new_finding(
        self,
        finding: contrast_fireball.AssessFinding,
        request: contrast_fireball.AssessRequest | None,
    ):
        contrast_fireball.new_finding(self.app_id, replace(finding, request=request))

    @_queue_if_app_uninitialized
    @_handle_errors()
    def new_libraries(self, libraries: list[contrast_fireball.Library]):
        """
        Record libraries that can be imported in the application.
        """
        contrast_fireball.new_libraries(self.app_id, libraries)

    @_queue_if_app_uninitialized
    @_handle_errors()
    def new_library_observations(
        self, observations: list[contrast_fireball.LibraryObservation]
    ):
        """
        Record observations of libraries imported in the application.
        Observations are reported periodically in batches.
        """
        contrast_fireball.new_library_observations(self.app_id, observations)

    @_queue_if_app_uninitialized
    @_handle_errors()
    def new_inventory_components(
        self, components: list[contrast_fireball.InventoryComponent]
    ):
        """
        Record Inventory Components.

        Components are reported periodically in batches. Duplicate items between sends
        will be ignored.
        """
        contrast_fireball.new_inventory_components(self.app_id, components)

    @contextlib.contextmanager
    def observability_trace(
        self,
        *,
        send_trace: bool,
        attributes: contrast_fireball.OtelAttributes | None = None,
    ) -> Generator[ObservabilityTrace | None, None, None]:
        """
        Manages the lifecycle of the root span for observability mode.

        `send_trace` determines whether or not the trace will actually be reported to
        the UI. In the future, we may decide not to send every trace, depending on
        observabilty sampling settings. Traces for requests containing an attack
        (identified by protect) will likely be sent regardless. For now, `send_trace`
        must be given an initial value, but this can be changed later by modifying the
        corresponding attribute on the trace object returned by this contextmanager.
        """
        # A new trace should only be created when a new request comes in to the server.
        # By this point, we must have called `initialize_application`, which sets these
        # critical attributes.
        assert hasattr(self, "app_id")
        assert self.config is not None

        if (
            trace_id := self._start_trace(
                contrast_fireball.SpanType.HttpServerRequest, attributes or {}
            )
        ) is None:
            logger.debug("No trace_id from fireball - not entering root span")
            yield None
            return

        trace = ObservabilityTrace(
            trace_id=trace_id, send_trace=send_trace, client=self
        )
        try:
            yield trace
        finally:
            self._end_trace(trace_id, trace.send_trace)
            logger.debug(
                "Trace sent to Fireball", trace_id=trace_id, send_trace=trace.send_trace
            )

    @_handle_errors()
    def _start_trace(
        self,
        action_type: contrast_fireball.SpanType,
        attributes: contrast_fireball.OtelAttributes,
    ) -> str | None:
        # the undecorated function can only return `str`, but the decorated function
        # returns `None` on error
        params = contrast_fireball.StartTraceParams(
            action_type=action_type,
            attributes=attributes,
        )
        result = contrast_fireball.start_trace(self.app_id, params)
        return result.data.trace_id

    @_handle_errors()
    def _end_trace(self, trace_id: str, send_trace: bool) -> None:
        params = contrast_fireball.EndTraceParams(
            trace_id=trace_id,
            send_trace=send_trace,
        )
        contrast_fireball.end_trace(self.app_id, params)

    @_handle_errors()
    def _update_trace(
        self, trace_id: str, attributes: contrast_fireball.OtelAttributes
    ) -> None:
        params = contrast_fireball.UpdateTraceParams(
            trace_id=trace_id, attributes=attributes
        )
        contrast_fireball.update_trace(self.app_id, params)

    @_handle_errors()
    def _get_trace_info(self, trace_id: str) -> contrast_fireball.TraceInfo | None:
        result = contrast_fireball.get_trace_info(self.app_id, trace_id)
        return result.data

    @_handle_errors()
    def _start_child_span(
        self,
        trace_id: str,
        action_type: contrast_fireball.SpanType,
        attributes: contrast_fireball.OtelAttributes,
        parent_span_id: str | None = None,
    ) -> str | None:
        # the undecorated function can only return `str`, but the decorated function
        # returns `None` on error
        params = contrast_fireball.StartChildSpanParams(
            trace_id=trace_id,
            action_type=action_type,
            attributes=attributes,
            parent_span_id=parent_span_id,
        )
        result = contrast_fireball.start_child_span(self.app_id, params)
        return result.data.id

    @_handle_errors()
    def _end_child_span(self, trace_id: str, span_id: str) -> None:
        params = contrast_fireball.EndChildSpanParams(
            trace_id=trace_id,
            span_id=span_id,
        )
        contrast_fireball.end_child_span(self.app_id, params)

    @_handle_errors()
    def _update_child_span(
        self, trace_id: str, span_id: str, attributes: contrast_fireball.OtelAttributes
    ) -> None:
        params = contrast_fireball.UpdateChildSpanParams(
            trace_id=trace_id,
            span_id=span_id,
            attributes=attributes,
        )
        contrast_fireball.update_child_span(self.app_id, params)


def agent_config_to_plain_dict(config: AgentConfig):
    """
    Convert all set options in the AgentConfig to a plain dictionary.
    """

    def conv(obj: object):
        if isinstance(obj, Enum):
            return obj.name
        return str(obj)

    json_config = {
        key: conv(v)
        for key, opt in config._config.items()
        if opt.source() != DEFAULT_VALUE_SRC and (v := opt.value()) is not None
    }

    # PROD-1745: Make sure to add the artifact hash to the session metadata.
    # This is implemented in get_session_metadata instead of as the
    # ConfigOption.default for application.session_metadata, because we always
    # want to send the artifact hash even if the user has set other session
    # metadata.
    if "application.session_id" not in json_config:
        json_config["application.session_metadata"] = config.get_session_metadata()

    return json_config


class ObservabilityTrace:
    trace_id: Final[str]
    send_trace: bool
    _client: Client

    def __init__(self, trace_id: str, send_trace: bool, client: Client):
        self.trace_id = trace_id
        self.send_trace = send_trace
        self._client = client

    def update(self, attributes: contrast_fireball.OtelAttributes) -> None:
        """
        Updates attributes on an existing trace.
        """
        self._client._update_trace(self.trace_id, attributes)

    def get_info(self) -> contrast_fireball.SpanInfo | None:
        """
        Gets trace info for an unsent trace.

        This function will return None if the root span is not found. Traces are closed
        and unavailable once they are sent.
        """
        if (trace_info := self._client._get_trace_info(self.trace_id)) is None:
            return None
        return trace_info.root_span

    @contextlib.contextmanager
    def child_span(
        self,
        action_type: contrast_fireball.SpanType,
        *,
        attributes: contrast_fireball.OtelAttributes | None = None,
        parent_span_id: str | None = None,
    ) -> Generator[ChildSpan | None, None, None]:
        """
        Manages the lifecycle of a child span for observability mode. The newly created
        child span will be attached to the current trace.
        """
        if (
            span_id := self._client._start_child_span(
                self.trace_id, action_type, attributes or {}, parent_span_id
            )
        ) is None:
            yield None
            return

        child = ChildSpan(span_id, self)
        try:
            yield child
        finally:
            self._client._end_child_span(self.trace_id, span_id)


class ChildSpan:
    span_id: Final[str]
    _trace: ObservabilityTrace

    def __init__(self, span_id: str, trace: ObservabilityTrace):
        self.span_id = span_id
        self._trace = trace

    def update(self, attributes: contrast_fireball.OtelAttributes) -> None:
        """
        Updates a child span with new attributes.
        """
        self._trace._client._update_child_span(
            self._trace.trace_id, self.span_id, attributes
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import re

from contrast.agent.agent_lib.input_tracing import (
    InputAnalysisResult,
    check_cmd_injection_query,
)
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.protect.rule.mode import Mode
from contrast.api.attack import Attack
from contrast.api.sample import Sample
from contrast.utils.loggers.logger import security_log_attack
from .base_rule import AnalysisStage, BaseRule

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class CmdInjection(BaseRule):
    """
    Command Injection Protection rule
    """

    RULE_NAME = "cmd-injection"

    def find_attack(
        self,
        candidate_string: str | None = None,
        analysis_stage: AnalysisStage | None = None,
        **kwargs,
    ) -> Attack | None:
        command_string = str(candidate_string) if candidate_string else None

        return super().find_attack(
            command_string, analysis_stage=analysis_stage, **kwargs
        )

    def build_attack_with_match(
        self,
        candidate_string: str | None,
        evaluation: InputAnalysisResult | None = None,
        attack: Attack | None = None,
        **kwargs,
    ) -> Attack | None:
        assert evaluation is not None
        assert candidate_string is not None
        for match in re.finditer(
            re.compile(re.escape(evaluation.input.value)), candidate_string
        ):
            input_len = match.end() - match.start()
            agent_lib_evaluation = check_cmd_injection_query(
                match.start(), input_len, candidate_string
            )
            if not agent_lib_evaluation:
                continue

            evaluation.attack_count += 1

            kwargs["startIndex"] = match.start()
            kwargs["endIndex"] = match.end()
            attack = self.build_or_append_attack(
                evaluation, attack, candidate_string, **kwargs
            )

        if attack is not None:
            attack.response = self.response_from_mode(self.mode)
            security_log_attack(attack, evaluation)

        return attack

    def build_attack_without_match(
        self,
        evaluation: InputAnalysisResult | None = None,
        attack: Attack | None = None,
        **kwargs,
    ) -> Attack | None:
        if self.mode == Mode.BLOCK_AT_PERIMETER:
            return super().build_attack_without_match(evaluation, attack, **kwargs)
        assert evaluation is not None
        if evaluation.score < 10:
            return None

        return next(
            (
                super(CmdInjection, self).build_attack_without_match(
                    full_eval, attack, **kwargs
                )
                for full_eval in evaluation.fully_evaluate()
                if full_eval.score >= 90
            ),
            None,
        )

    def build_sample(
        self,
        evaluation: InputAnalysisResult | None,
        candidate_string: str | None,
        **kwargs,
    ) -> Sample:
        assert evaluation is not None
        sample = self.build_base_sample(evaluation)
        command = candidate_string
        if command is not None:
            sample.details["command"] = command

        if "startIndex" in kwargs:
            sample.details["startIndex"] = int(kwargs["startIndex"])

        if "endIndex" in kwargs:
            sample.details["endIndex"] = int(kwargs["endIndex"])

        return sample

    def infilter_kwargs(
        self, user_input: str, patch_policy: PatchLocationPolicy
    ) -> dict[str, object]:
        return dict(method=patch_policy.method_name, original_command=user_input)

    def skip_protect_analysis(
        self, user_input: object, args: tuple, kwargs: dict[str, object]
    ) -> bool:
        """
        cmdi rule supports list user input as well as str and bytes
        Do not skip protect analysis if user input is a  populated list
        """
        if isinstance(user_input, list) and user_input:
            return False

        return super().skip_protect_analysis(user_input, args, kwargs)

    def convert_input(self, user_input: object) -> str:
        if isinstance(user_input, list):
            user_input = " ".join(user_input)

        return super().convert_input(user_input)

    def _infilter(self, match_string: str, **kwargs):
        # TODO: PYT-3088
        #  deserialization_rule = Settings().protect_rules[Deserialization.RULE_NAME]
        #  deserialization_rule.check_for_deserialization()

        super()._infilter(match_string, **kwargs)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast_vendor.stdlib_list import stdlib_list

# this logic (and all of vendored isort) can be removed when we drop py39
version_info = sys.version_info
_stdlib_modules = (
    set(stdlib_list("3.9")) if version_info[:2] == (3, 9) else sys.stdlib_module_names
)


def is_stdlib_module(module_name):
    """
    Returns True if module_name belongs to standard library module, False otherwise.

    NOTE: 'test' is included in _stdlib_modules so if we're testing this,
    we cannot pass in a module that starts with test.file...
    """
    top_module_name = module_name.split(".")[0]
    return top_module_name in _stdlib_modules



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import copy
from itertools import product


def _to_list(obj):
    return obj if isinstance(obj, list) else [obj]


def _generate_nodes(node_cls, node_kwargs, dataflow=None):
    """
    Generates policy objects using the given node class and kwargs

    The "module", "method_name", and "class_name" attributes can optionally be lists.
    This function generates nodes based on the product of those attributes, which
    helps reduce code duplication in policy.
    """
    node_kwargs = copy.copy(node_kwargs)
    # NOTE: this really needs to be fixed on the nodes themselves
    node_kwargs.setdefault("instance_method", True)

    if dataflow is not None:
        node_kwargs["dataflow"] = dataflow

    modules = _to_list(node_kwargs.pop("module"))
    method_names = _to_list(node_kwargs.pop("method_name"))
    class_names = _to_list(node_kwargs.pop("class_name", ""))

    for module, class_name, method_name in product(modules, class_names, method_names):
        yield node_cls(
            module=module,
            class_name=class_name,
            method_name=method_name,
            **node_kwargs,
        )


def generate_policy(node_cls, nodes, dataflow=None):
    """
    Generate policy objects given a policy node class and a list of putative nodes

    Objects in the node list can either by Python dicts or CompositeNode instances.
    """
    for node in nodes:
        yield from (
            node.generate_from_class(node_cls, dataflow=dataflow)
            if isinstance(node, CompositeNode)
            else _generate_nodes(node_cls, node, dataflow=dataflow)
        )


class CompositeNode:
    """
    CompositeNode allows multiple policy nodes to share an arbitrary set of attributes

    This is intended to reduce duplication in policy definitions by reducing
    the number of repeated attributes.
    """

    def __init__(self, common_kwargs, nodes):
        self.common_kwargs = common_kwargs
        self.nodes = nodes

    def generate_from_class(self, node_cls, dataflow=None):
        """
        Generates policy objects of the given class based on common kwargs and the list of nodes
        """
        for node in self.nodes:
            # NOTE: the order of composition matters here in order to allow
            # more specific nodes to override the common kwargs
            # Once we support only >=3.9 we can use the | operator for dicts
            kwargs = {**self.common_kwargs, **node}
            yield from _generate_nodes(node_cls, kwargs, dataflow=dataflow)


def get_self_for_method(patch_policy, args):
    """
    Retrieves self for a method's PatchLocationPolicy,

    If any node in the policy has a False instance_method attribute return None
    """
    for node in patch_policy.all_nodes:
        if not node.instance_method:
            return None

    return args[0] if args else None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import collections
import functools
import hashlib
import itertools
import os
import re
from types import ModuleType

from contrast_vendor.importlib_metadata import (
    Distribution,
    distributions as importlib_distributions,
    _top_level_declared,
    _top_level_inferred,
)
from contrast_vendor.importlib_metadata._meta import PackageMetadata


@functools.cache
def distributions() -> list[Distribution]:
    """
    Get all Distribution instances for the current environment.
    """
    return list(importlib_distributions())


@functools.cache
def packages_distributions() -> dict[str, list[Distribution]]:
    """
    Return a mapping of top-level packages to their distributions.
    """
    pkg_to_dist = collections.defaultdict(list)
    for dist in distributions():
        for pkg in _top_level_declared(dist) or _top_level_inferred(dist):
            if dist not in pkg_to_dist[pkg]:
                pkg_to_dist[pkg].append(dist)
    return dict(pkg_to_dist)


@functools.cache
def namespace_file_to_distribution() -> dict[str, Distribution]:
    namespacepkg_dists = [
        dists for pkg, dists in packages_distributions().items() if len(dists) > 1
    ]
    files_to_dist = {
        str(f): dist
        for dist in itertools.chain.from_iterable(namespacepkg_dists)
        for f in dist.files or []
    }
    return files_to_dist


def get_module_distribution_metadata(module: ModuleType) -> PackageMetadata | None:
    top_level_name = module.__name__.partition(".")[0]
    if dist_meta := _get_simple_regular_package_distribution_metadata(top_level_name):
        return dist_meta
    elif (
        (mod_file := getattr(module, "__file__", None))
        and (file_name := normalize_file_name(mod_file))
        and (dist := namespace_file_to_distribution().get(file_name))
    ):
        return dist.metadata
    return None


@functools.lru_cache
def _get_simple_regular_package_distribution_metadata(
    package_name: str,
) -> PackageMetadata | None:
    """
    Get the distribution metadata for a package that has only one distribution.

    This handles the common case of a regular package (https://docs.python.org/3/glossary.html#term-regular-package).

    This function is separated from get_module_distribution_metadata to allow for caching.
    For packages with many modules, this caching makes a measurable difference in performance
    because Distribution.metadata is a relatively expensive operation, reading from the filesystem
    and parsing files with the standard library's email.parser module.
    """
    dists = packages_distributions().get(package_name)
    if not dists:
        return None
    if len(dists) == 1:
        return dists[0].metadata
    return None


@functools.cache
def get_installed_dist_names():
    return [dist.name for dist in distributions()]


SITE_PACKAGES_DIR = f"{os.sep}site-packages{os.sep}"
DIST_PACKAGES_DIR = f"{os.sep}dist-packages{os.sep}"


def normalize_file_name(file_path: str) -> str | None:
    """
    This function normalizes the file path by removing the leading site-packages or dist-packages portion
    and removing the .pyc suffix if present, returning the resulting .py file path.
    @param file_path: the full path to a Python .pyc or .py file
    @return: the normalized file path ending in .py, or None if prefix not found
    """
    _, match, normalized_file_name = file_path.rpartition(SITE_PACKAGES_DIR)
    if not match:
        _, match, normalized_file_name = file_path.rpartition(DIST_PACKAGES_DIR)
        if not match:
            return None

    if normalized_file_name.endswith(".pyc"):
        normalized_file_name = normalized_file_name.removesuffix("c")
    return normalized_file_name


def normalize_dist_name(name):
    # Normalize the name based on the spec https://packaging.python.org/en/latest/specifications/name-normalization/
    return re.sub(r"[-_.]+", "-", name).lower()


def get_hash(name, version):
    """
    DO NOT ALTER OR REMOVE

    This must match the calculation made by the artifact dependency management database.
    """
    name = normalize_dist_name(name)

    to_hash = name + " " + version

    return hashlib.sha1(to_hash.encode("utf-8")).hexdigest()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import base64
import json

from contrast.utils.configuration_utils import str_to_bool
from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast import AGENT_CURR_WORKING_DIR


from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")


class Api(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(canonical_name="api.url", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="api.service_key",
                default_value="",
                type_cast=str,
                redacted=True,
            ),
            ConfigOption(
                canonical_name="api.api_key",
                default_value="",
                type_cast=str,
                redacted=True,
            ),
            ConfigOption(
                canonical_name="api.user_name",
                default_value="",
                type_cast=str,
                redacted=True,
            ),
            ConfigOption(
                canonical_name="api.token",
                default_value="",
                type_cast=str,
                redacted=True,
            ),
            ConfigOption(
                canonical_name="api.reporting_client",
                default_value="fireball",
                type_cast=str,
                log_effective_config=False,
            ),
            ConfigOption(
                canonical_name="api.request_audit.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="api.request_audit.path",
                default_value=AGENT_CURR_WORKING_DIR,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="api.request_audit.requests",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="api.request_audit.responses",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="api.certificate.enable",
                default_value=True,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="api.certificate.ignore_cert_errors",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="api.certificate.ca_file",
                default_value="",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="api.certificate.cert_file",
                default_value="",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="api.certificate.key_file",
                default_value="",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="api.proxy.enable",
                default_value=False,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="api.proxy.url",
                default_value="",
                type_cast=str,
                redacted=True,
            ),
        ]

    def build(self, config: dict, yaml_config: list[tuple[dict, str]]):
        api_keys = {
            "api.url": "url",
            "api.service_key": "service_key",  # pragma: allowlist secret
            "api.api_key": "api_key",  # pragma: allowlist secret
            "api.user_name": "user_name",
        }

        error = super().build(config, yaml_config)

        if error:
            return error

        api_token = config.get("api.token")

        if not (token_value := api_token.value()):
            api_token.log_effective_config = False
            return None

        try:
            api_values = json.loads(base64.b64decode(token_value, validate=True))
        except Exception as e:
            error = f"Invalid value on {api_token.canonical_name} - {e}"
            return error

        for api_config_key, api_config_key_teamserver in api_keys.items():
            option = config.get(api_config_key)
            if not option.is_definitely_static():
                if (
                    new_value := api_values.get(api_config_key_teamserver, None)
                ) is None:
                    error = f"Invalid value on {api_token.canonical_name}"
                    logger.error(
                        "Unable to set %s based on %s",
                        option.canonical_name,
                        api_token.canonical_name,
                    )
                    break

                option.override_value = new_value
                option.log_effective_config = False
            else:
                logger.info(
                    f"Using configured value for {option.canonical_name} set using {option.source()} instead of api.token"
                )

        return error



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging
from contrast_fireball import Library

logger = logging.getLogger("contrast")


class ApplicationUpdate(BaseTsAppMessage):
    def __init__(self, libraries: list[Library]):
        super().__init__()

        self.extra_headers["Content-Type"] = "application/json"

        # activity message sends "components" aka "architectures"
        # so we will not send the "components" field at this time.

        # field "timestamp" represents the amount of time that has passed
        # since the app settings were changed (not an actual timestamp).
        self.body = {
            "timestamp": self.since_last_update,
            "libraries": [
                {
                    "classCount": lib.class_count,
                    "file": lib.file,
                    "hash": lib.hash,
                    "url": lib.url,
                    "version": lib.version,
                    "externalDate": lib.external_date,
                    "internalDate": lib.internal_date,
                    "tags": self.settings.config.get("inventory.tags"),
                }
                for lib in libraries
                if lib.hash
            ],
        }

    @property
    def name(self):
        return "application-update"

    @property
    def path(self):
        return "update/application"

    @property
    def request_method(self):
        return requests.put

    @fail_loudly("Failed to process ApplicationUpdate response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from functools import wraps
from contrast_vendor import wrapt

from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.policy.analysis import analyze
from contrast.utils.patch_utils import add_watermark
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_assess_method_legacy(original_method, patch_policy):
    """
    Creates method to replace old method and call our assess code with the original method

    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :return: Newly created patch function
    """

    def assess_method(*args, **kwargs):
        result = None

        try:
            result = original_method(*args, **kwargs)
        finally:
            analyze(patch_policy, result, args, kwargs)

        return result

    return add_watermark(assess_method)


def build_assess_method(original_method, patch_policy):
    """
    identical to build_assess_method, but with a wrapt wrapper

    From the wrapt documentation:
        In all cases, the wrapped function passed to the wrapper function is called
        in the same way, with args and kwargs being passed. The instance argument
        doesn't need to be used in calling the wrapped function.
    """

    @wrapt.function_wrapper
    def assess_method_wrapper(wrapped, instance, args, kwargs):
        result = None

        try:
            result = wrapped(*args, **kwargs)
        finally:
            if instance is not None:
                args = (instance, *args)
            analyze(patch_policy, result, args, kwargs)

        return result

    # note that this function likely already has our watermark
    return add_watermark(assess_method_wrapper(original_method))


def build_assess_async_method(original_method, patch_policy):
    """
    Build a generic async method which instruments original_method.

    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :return: Newly created async patch function
    """

    async def assess_method(*args, **kwargs):
        result = None

        try:
            result = await original_method(*args, **kwargs)
        finally:
            analyze(patch_policy, result, args, kwargs)

        return result

    return add_watermark(assess_method)


def build_assess_classmethod(original_method, patch_policy):
    """
    Creates method to replace old method and call our assess code with the original method

    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :return: Newly created patch function

    A separate method was required for classmethod patch because we need to remove
    argument 1. arg 1 is the class. This is something that is automatically passed to
    the function so passing it again will cause a TypeError.
    """
    original_method = patch_manager.as_func(original_method)

    def assess_classmethod(*args, **kwargs):
        result = None

        try:
            result = original_method(*args, **kwargs)
        finally:
            analyze(patch_policy, result, args, kwargs)

        return result

    return add_watermark(assess_classmethod)


def build_assess_deadzone(original_method, patch_policy):
    """
    Creates patch method that calls original method in contrast scope

    This prevents any analysis down the stack.

    :param original_method: method to call for result
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :return: Newly created patch function
    """

    @wraps(original_method)
    def assess_deadzone(*args, **kwargs):
        with scope.contrast_scope():
            return original_method(*args, **kwargs)

    return add_watermark(assess_deadzone)


def build_assess_async_deadzone(original_method, patch_policy):
    """
    The equivalent of `build_assess_deadzone`, but for async
    """

    async def assess_deadzone(*args, **kwargs):
        with scope.contrast_scope():
            return await original_method(*args, **kwargs)

    return add_watermark(assess_deadzone)


def build_assess_property_fget(orig_property, patch_policy):
    """
    Creates property getter to replace old property and call assess code for analysis

    The new property calls the original property and then runs assess analysis.
    """

    def assess_property(*args, **kwargs):
        result = None

        try:
            instance = args[0]
            result = orig_property.__get__(instance)
        finally:
            analyze(patch_policy, result, args, kwargs)

        return result

    return add_watermark(assess_property)


def apply_cached_property(cls_or_module, patch_policy, property_name, orig_property):
    """
    Older werkzeug versions implement cached_property that does not inherit from property.
    This causes us to have to use a workaround for patching to avoid errors.
    Instead of replacing the cached_property with a new property, we replace it with
    and object proxy with a custom __get__ method.
    """
    proxied_property = WerkzeugCachedPropertyProxy(
        orig_property, property_name, patch_policy
    )

    try:
        setattr(cls_or_module, property_name, proxied_property)
    except Exception as ex:
        logger.debug(
            "Failed to apply patch to cached_property: %s", property_name, exc_info=ex
        )

    return True


class WerkzeugCachedPropertyProxy(wrapt.ObjectProxy):
    cs__attr_name = None
    cs__patch_policy = None

    def __init__(self, wrapped, attr_name, patch_policy):
        super().__init__(wrapped)
        self.cs__patch_policy = patch_policy
        self.cs__attr_name = attr_name

    def __get__(__cs_self, *args, **kwargs):
        result = __cs_self.__wrapped__.__get__(*args, **kwargs)

        try:
            # Self is the only arg that seems to be relevant for policy/reporting
            args = (__cs_self.__wrapped__,)
            analyze(__cs_self.cs__patch_policy, result, args, {})
        except Exception as ex:
            logger.debug(
                "Failed to apply policy for %s", __cs_self.cs__attr_name, exc_info=ex
            )

        return result



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from contrast.agent.agent_lib.input_tracing import InputAnalysisResult
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.protect.rule.base_rule import AnalysisStage, BaseRule
from contrast.agent.protect.rule.deserialization.custom_searcher import CustomSearcher
from contrast.agent.protect.rule.deserialization.pickle_searcher import PickleSearcher
from contrast.agent.protect.rule.deserialization.yaml_searcher import YAMLSearcher
from contrast.api import user_input
from contrast.api.attack import Attack
from contrast.api.sample import Sample
from contrast.utils.string_utils import ends_with_any
from contrast.utils.stack_trace_utils import build_and_clean_stack

from contrast_vendor import structlog as logging
import contextlib

logger = logging.getLogger("contrast")


class Deserialization(BaseRule):
    """
    Deserialization Protection rule
    """

    RULE_NAME = "untrusted-deserialization"

    # pickle and pyyaml both use load
    METHODS = [
        "loads",
        "load",
        "construct_object",
        "construct_python_object_apply",
        "construct_mapping",
        "make_python_instance",
    ]
    FILENAMES = ["pickle.py", "yaml.constructor.py", "yaml.__init__.py"]

    UNKNOWN = "UNKNOWN"

    @property
    def custom_searchers(self) -> list[CustomSearcher]:
        return [PickleSearcher(), YAMLSearcher()]

    def is_prefilter(self) -> bool:
        return False

    def is_postfilter(self) -> bool:
        return False

    def skip_protect_analysis(
        self, user_input: object, args: tuple, kwargs: dict[str, object]
    ) -> bool:
        """
        Deserialization rule will receive io streams as user input.

        :return: Bool if to skip running protect infilter
        """
        if not user_input:
            return True

        # checking if obj has attr "read" is more robust than using isinstance
        if hasattr(user_input, "read"):
            return False

        return super().skip_protect_analysis(user_input, args, kwargs)

    def convert_input(self, user_input: object) -> str:
        if isinstance(user_input, (str, bytes)):
            data = user_input
        else:
            data = self._get_stream_data(user_input)

        return super().convert_input(data)

    def _get_stream_data(self, user_input: object) -> str:
        """
        Get data from a stream object but make sure to return the stream position
        to the original location.

        :param user_input: obj we expect to be a stream with attrs read, tell and seek
        :return: str or bytes
        """
        if not all(hasattr(user_input, attr) for attr in ["read", "tell", "seek"]):
            return ""

        # Find current steam position
        try:
            seek_loc = user_input.tell()
        except Exception:
            seek_loc = 0

        # Read the object data
        try:
            data = user_input.read()
        except Exception:
            data = ""

        # Return object to original stream position so it can be re-read
        with contextlib.suppress(Exception):
            user_input.seek(seek_loc)

        return data

    def find_attack(
        self,
        candidate_string: str | None = None,
        analysis_stage: AnalysisStage | None = None,
        **kwargs,
    ) -> Attack | None:
        """
        Finds the attacker in the original string if present
        """
        assert candidate_string is not None
        logger.debug("Checking for %s in %s", self.name, candidate_string)

        attack = None
        if self.evaluate_custom_searchers(candidate_string):
            evaluation = self.build_evaluation(candidate_string)
            attack = self.build_attack_with_match(
                candidate_string, evaluation, attack, **kwargs
            )

        return attack

    def report_attack_without_finding(self, value, **kwargs):
        if attack := self.build_attack_with_match(value, None, **kwargs):
            self._append_to_context(attack)

    def check_for_deserialization(self):
        """
        For the sandbox feature of this rule, we need to determine if we're in a deserializer when a command is called.
        Command injection's infilter method should call this to check and let us handle attack detection and exception
        raising before doing their work.
        """
        found_on_stack = False

        # TODO: PYT-3088 get the stack_elements
        stack_elements = []
        for element in stack_elements[::-1]:
            lower_file_name = element.file_name.lower()

            if (
                element.method_name
                and element.method_name in self.METHODS
                and (
                    lower_file_name in self.FILENAMES
                    or ends_with_any(lower_file_name, self.FILENAMES)
                )
            ):
                found_on_stack = True
                break

        # TODO: PYT-3088
        #  determine what value to pass here; modify this method signature as needed
        if found_on_stack:
            pass
        #     self.report_attack_without_finding("")
        #     if self.should_block(attack):
        #         raise contrast.SecurityException(rule_name=self.name)

    def build_sample(
        self,
        evaluation: InputAnalysisResult | None,
        candidate_string: str | None,
        **kwargs,
    ) -> Sample:
        assert evaluation is not None
        sample = self.build_base_sample(evaluation)

        sample.details["command"] = False

        if "deserializer" in kwargs:
            sample.details["deserializer"] = kwargs["deserializer"]

        return sample

    def evaluate_custom_searchers(self, attack_vector: str):
        searcher_score = 0
        for searcher in self.custom_searchers:
            impact = searcher.impact_of(attack_vector)

            if impact > 0:
                logger.debug("Match on custom searcher: %s", searcher.searcher_id)

                searcher_score += impact
                if searcher_score >= searcher.IMPACT_HIGH:
                    return True

        return False

    def build_evaluation(self, value: str) -> InputAnalysisResult:
        """
        Given a user-input value, aka gadget, create an InputAnalysisResult instance.

        :param value: the user input containing a Gadget
        """
        return InputAnalysisResult(
            user_input.UserInput(type=user_input.InputType.UNKNOWN, value=value),
            self.RULE_NAME,
            0,
        )

    def infilter_kwargs(self, user_input: str, patch_policy: PatchLocationPolicy):
        stack_elements = build_and_clean_stack()

        return dict(
            deserializer=patch_policy.method_name, stack_elements=stack_elements
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)


def build_noop_patch(original_func, patch_policy):
    del patch_policy

    def noop_intern(wrapped, instance, args, kwargs):
        """
        Defeat interning by returning the original string

        sys.intern takes no kwargs
        """
        del wrapped, instance, kwargs

        return args[0]

    return wrap_and_watermark(original_func, noop_intern)


def patch_sys(module):
    build_and_apply_patch(module, "intern", build_noop_patch)


def register_patches():
    register_module_patcher(patch_sys, "sys")


def reverse_patches():
    unregister_module_patcher("sys")
    # sys is always imported, no need to check
    patch_manager.reverse_patches_by_owner(sys)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

from functools import cached_property
from typing import Literal
from typing import TYPE_CHECKING
import contrast
from contrast.agent import scope
from contrast.agent.agent_lib.input_tracing import (
    InputAnalysisResult,
)
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.settings import Settings
from contrast.api.attack import Attack, ProtectResponse
from contrast.api.sample import Sample
from contrast.agent.protect.rule.mode import Mode
from contrast.api.user_input import UserInput
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast.utils.stack_trace_utils import StackSummary, build_and_clean_stack
from contrast_vendor import structlog as logging
from contrast.utils.loggers.logger import security_log_attack
from contrast.utils.string_utils import ensure_string

if TYPE_CHECKING:
    from contrast.agent.request_context import RequestContext


logger = logging.getLogger("contrast")


BLOCKING_RULES = frozenset([Mode.BLOCK, Mode.BLOCK_AT_PERIMETER])
PREFILTER_RULES = frozenset([Mode.BLOCK_AT_PERIMETER])
POSTFILTER_RULES = frozenset([Mode.BLOCK, Mode.MONITOR])

PREFILTER = "prefilter"
INFILTER = "infilter"
POSTFILTER = "postfilter"
AnalysisStage = Literal["prefilter", "infilter", "postfilter"]


class BaseRule:
    """
    Base rule object that all protection rules will inherit
    """

    RULE_NAME = "base-rule"

    def __init__(self):
        self.settings = Settings()
        self.settings.protect_rules[self.name] = self
        self.probe_analysis_enabled = self.settings.config.get(
            "protect.probe_analysis.enable"
        )

    @property
    def name(self):
        return self.RULE_NAME

    @property
    def mode(self) -> Mode:
        return self.settings.config.get(self.config_rule_path_mode, Mode.OFF)

    @cached_property
    def config_rule_path_mode(self) -> str:
        return f"protect.rules.{self.name}.mode"

    def is_prefilter(self) -> bool:
        """
        Checks if a rules mode is for prefilter
        """
        return self.enabled and self.mode in PREFILTER_RULES

    def is_postfilter(self) -> bool:
        """
        Checks if a rules mode is for postfilter
        """
        return self.enabled and self.mode in POSTFILTER_RULES

    def is_blocked(self) -> bool:
        """
        Checks if a rules mode is for blocking
        """
        return self.enabled and self.mode in BLOCKING_RULES

    @property
    def enabled(self) -> bool:
        """
        A rule is enabled only if all 3 conditions are met:
        1. rule is not in disabled rules list
        2. rule mode is not OFF
        3. an exclusion wasn't applied from Teamserver
        """
        disabled_rules = self.settings.config.get("protect.rules.disabled_rules")
        if disabled_rules and self.name in disabled_rules:
            return False

        req_ctx = contrast.REQUEST_CONTEXT.get()
        if req_ctx is not None and req_ctx.excluded_protect_rules:
            return self.name not in req_ctx.excluded_protect_rules

        return self.mode != Mode.OFF

    def should_block(self, attack: Attack | None) -> bool:
        return bool(attack and attack.response == ProtectResponse.BLOCKED)

    def prefilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Will throw a SecurityException if a response needs to be blocked
        """
        logger.debug("PROTECT: Prefilter for %s", self.name)

        attack = self.find_attack(analysis_stage=PREFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if attack.response == ProtectResponse.BLOCKED_AT_PERIMETER:
            raise contrast.SecurityException(rule_name=self.name)

    def _infilter(self, match_string: str, **kwargs):
        """
        Scans the input analysis for the rule and looks for matched attack signatures. The call to this method may be
        rule specific and include additional context in a args list.
        """
        if self.mode == Mode.OFF:
            return

        logger.debug("PROTECT: Infilter for %s", self.name)

        attack = self.find_attack(match_string, analysis_stage=INFILTER, **kwargs)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise contrast.SecurityException(rule_name=self.name)

    @fail_loudly("Failed to run protect rule")
    def protect(
        self,
        patch_policy: PatchLocationPolicy,
        user_input: object,
        args: tuple,
        kwargs: dict[str, object],
    ):
        if not self.enabled:
            return

        if self.skip_protect_analysis(user_input, args, kwargs):
            return

        with scope.contrast_scope():
            user_input = self.convert_input(user_input)
            if not user_input:
                return

            self.log_safely(patch_policy.method_name, user_input)

            self._infilter(user_input, **self.infilter_kwargs(user_input, patch_policy))

    def infilter_kwargs(self, user_input: str, patch_policy: PatchLocationPolicy):
        return {}

    def skip_protect_analysis(
        self, user_input: object, args: tuple, kwargs: dict[str, object]
    ) -> bool:
        """
        We only want to run protect on user input that is of a type supported
        by the rule.

        Most rules use this implementation, but some override this depending on
        expected user input types.

        :return: Bool if to skip running protect infilter
        """
        if not user_input:
            return True

        if isinstance(user_input, (str, bytes)):
            return False

        logger.debug(
            "WARNING: unknown input type %s for rule %s", type(user_input), self.name
        )

        return True

    def convert_input(self, user_input: object) -> str:
        return ensure_string(user_input)

    def postfilter(self):
        """
        Scans the input analysis for the rule and looks for matched attack signatures

        Appends attacker to the context if a positive evaluation is found
        """
        if self.mode == Mode.OFF or not self.probe_analysis_enabled:
            return

        logger.debug("PROTECT: Postfilter", rule=self.name)

        attack = self.find_attack(analysis_stage=POSTFILTER)
        if attack is None or len(attack.samples) == 0:
            return

        self._append_to_context(attack)

        if self.should_block(attack):
            raise contrast.SecurityException(rule_name=self.name)

    def find_attack(
        self,
        candidate_string: str | None = None,
        analysis_stage: AnalysisStage | None = None,
        **kwargs,
    ) -> Attack | None:
        """
        Finds the attacker in the original string if present
        """
        if candidate_string is not None:
            logger.debug("Checking for %s in %s", self.name, candidate_string)

        evaluations_for_rule = self.evaluations_for_rule()

        attack = None
        for evaluation in evaluations_for_rule:
            if analysis_stage == POSTFILTER and evaluation.attack_count > 0:
                continue
            if candidate_string:
                if candidate_string.find(evaluation.input.value) == -1:
                    continue

                attack = self.build_attack_with_match(
                    candidate_string, evaluation, attack, **kwargs
                )
            else:
                attack = self.build_attack_without_match(evaluation, attack, **kwargs)

        return attack

    def build_attack_with_match(
        self,
        candidate_string: str | None,
        evaluation: InputAnalysisResult | None = None,
        attack: Attack | None = None,
        **kwargs,
    ) -> Attack | None:
        attack = self.build_or_append_attack(
            evaluation, attack, candidate_string, **kwargs
        )

        if evaluation:
            evaluation.attack_count += 1

        attack.response = self.response_from_mode(self.mode)
        security_log_attack(attack, evaluation)
        return attack

    def build_attack_without_match(
        self,
        evaluation: InputAnalysisResult | None = None,
        attack: Attack | None = None,
        **kwargs,
    ) -> Attack | None:
        if evaluation and evaluation.score < 10:
            return None
        if self.mode == Mode.BLOCK_AT_PERIMETER:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)

            attack.response = self.response_from_mode(self.mode)
            security_log_attack(attack, evaluation)
        elif evaluation is None or evaluation.attack_count == 0:
            attack = self.build_or_append_attack(evaluation, attack, **kwargs)
            attack.response = ProtectResponse.PROBED
            security_log_attack(attack, evaluation)

        return attack

    def build_or_append_attack(
        self,
        evaluation: InputAnalysisResult | None,
        attack: Attack | None = None,
        candidate_string: str | None = None,
        **kwargs,
    ) -> Attack:
        if attack is None:
            attack = self.build_base_attack()

        attack.add_sample(self.build_sample(evaluation, candidate_string, **kwargs))

        return attack

    def build_base_attack(self) -> Attack:
        return Attack(self.name)

    def build_sample(
        self,
        evaluation: InputAnalysisResult | None,
        candidate_string: str | None,
        **kwargs,
    ) -> Sample:
        assert evaluation is not None
        return self.build_base_sample(evaluation)

    def build_user_input(self, evaluation: InputAnalysisResult) -> UserInput:
        return evaluation.input

    def build_base_sample(
        self,
        evaluation: InputAnalysisResult,
        prebuilt_stack: StackSummary | None = None,
    ) -> Sample:
        return Sample(
            user_input=self.build_user_input(evaluation),
            stack=prebuilt_stack if prebuilt_stack else build_and_clean_stack(),
        )

    def _append_to_context(self, attack: Attack):
        context = contrast.REQUEST_CONTEXT.get()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug("WARNING: failed to get request context in _append_to_context")
            return

        context.attacks.append(attack)

    _RESPONSE_MAP = {
        Mode.MONITOR: ProtectResponse.MONITORED,
        Mode.BLOCK: ProtectResponse.BLOCKED,
        Mode.BLOCK_AT_PERIMETER: ProtectResponse.BLOCKED_AT_PERIMETER,
        Mode.OFF: ProtectResponse.NO_ACTION,
    }

    def response_from_mode(self, mode: Mode) -> ProtectResponse | None:
        return self._RESPONSE_MAP.get(mode)

    def evaluations_for_rule(
        self, context: RequestContext | None = None
    ) -> list[InputAnalysisResult]:
        if context is None:
            context = contrast.REQUEST_CONTEXT.get()
        if context is None:
            # do not remove; this case is not yet well-understood
            logger.debug(
                "WARNING: failed to get request context in evaluations_for_rule"
            )
            return []

        return [
            evaluation
            for evaluation in context.user_input_analysis
            if evaluation.rule_id == self.RULE_NAME
        ]

    @fail_quietly("Failed to log user input for protect rule")
    def log_safely(self, method_name: str, user_input: object):
        """
        Attempt to log user supplied input but do not fail if unable to do so.
        """
        logger.debug(
            "Applying %s rule method %s with user input %s",
            self.name,
            method_name,
            ensure_string(user_input, errors="replace"),
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.settings import Settings
from contrast.agent.assess.rules.static_rule import StaticRule


class BaseConfigRule(StaticRule):
    SESSION_ID = "sessionId"
    PATH = "path"
    SNIPPET = "snippet"

    def get_snippet(self, value):
        return NotImplementedError

    def create_properties(self, value, config_path):
        settings = Settings()

        properties = {}
        properties[self.SESSION_ID] = settings.config.session_id
        properties[self.PATH] = config_path
        properties[self.SNIPPET] = self.get_snippet(value)
        return properties



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.


class GlomDict(dict):
    """
    A dictionary that allows for dot notation access to nested dictionaries.

    Inspired by the glom library.
    """

    def __getitem__(self, key):
        work_dict = super()
        *parents, key = key.split(".")
        for parent in parents:
            work_dict = work_dict.__getitem__(parent)
        return work_dict.__getitem__(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, AttributeError):
            return default



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_CMD_INJECTION",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_CMD_INJECTION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "BASE64_ENCODED",
    "CSS_ENCODED",
    "CSV_ENCODED",
    "HTML_ENCODED",
    "JAVASCRIPT_ENCODED",
    "JAVA_ENCODED",
    "LDAP_ENCODED",
    "OS_ENCODED",
    "SQL_ENCODED",
    "URL_ENCODED",
    "VBSCRIPT_ENCODED",
    "XML_ENCODED",
    "XPATH_ENCODED",
]


cmd_injection_triggers = [
    {
        "module": "os",
        "method_name": "popen",
        "source": "ARG_0,KWARG:cmd",
        "protect_mode": True,
    },
    {
        "module": "os",
        "method_name": "system",
        "source": "ARG_0,KWARG:command",
        "protect_mode": True,
    },
    {
        "module": "subprocess",
        "method_name": ["call", "check_call", "check_output"],
        "source": "ARG_0,KWARG:args",
        "action": "SUBPROCESS",
        "protect_mode": True,
    },
    {
        "module": "subprocess",
        "class_name": "Popen",
        "method_name": "__init__",
        "instance_method": True,
        "source": "ARG_0,KWARG:args",
        "action": "SUBPROCESS",
        "protect_mode": True,
    },
    {
        "module": "os",
        "method_name": ["spawnv", "spawnvp", "spawnve", "spawnvpe"],
        "source": "ARG_1,KWARG:file,ARG_2,KWARG:args",
        "protect_mode": True,
    },
    {
        "module": "os",
        "method_name": ["execv", "execvp", "execve", "execvpe"],
        "source": "ARG_0,KWARG:file,ARG_1,KWARG:args",
        "protect_mode": True,
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "cmd-injection",
        cmd_injection_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
from typing import Callable
from urllib.parse import SplitResult, urlsplit, urlunsplit
import http.client

from contrast_fireball import OtelAttributes, SpanType

from contrast.agent.policy.handlers import EventDict
from contrast.utils.safe_import import safe_import


REDACTED_CREDENTIALS = "REDACTED:REDACTED@"


class _Sentinel: ...


yarl_URL = safe_import("yarl.URL") or _Sentinel


def _clean_url(parsed: SplitResult) -> SplitResult:
    """
    Performs rudimentary masking on the provided URL, redacting the username and
    password, and removing the querystring and fragment. Also removes the port if it is
    implied by the scheme.

    Operates on urllib SplitResult objects. Unlike `urlparse`, `urlsplit` does not have
    special handling for a rarely-used URL feature called "params", which can be applied
    to particular path segments using a semi-colon (;). We're not concerning ourselves
    with this; urlsplit includes those in `path`.
    """
    credentials = (
        REDACTED_CREDENTIALS
        if parsed.username is not None or parsed.password is not None
        else ""
    )

    port = ""
    port_implied = (
        parsed.port is None
        or (parsed.scheme == "http" and parsed.port == 80)
        or (parsed.scheme == "https" and parsed.port == 443)
    )
    if not port_implied:
        assert parsed.port is not None
        port = f":{parsed.port}"

    netloc = f"{credentials}{parsed.hostname or ''}{port}"

    return SplitResult(
        parsed.scheme,
        netloc,
        parsed.path,
        "",  # query
        "",  # fragment
    )


def _build_attrs(url: str) -> OtelAttributes:
    cleaned_parsed_url = _clean_url(urlsplit(url))

    if cleaned_parsed_url.hostname is None:
        raise ValueError(f"Observe failed to extract hostname: {url=}")

    attrs: OtelAttributes = {
        "server.address": cleaned_parsed_url.hostname,
        "url.full": urlunsplit(cleaned_parsed_url),
    }

    if (
        cleaned_parsed_url.port is not None
    ):  # implied ports (80, 443) are removed in _clean_url
        attrs["server.port"] = cleaned_parsed_url.port

    return attrs


def observe_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    url_param_location = event_dict.get("url", "")

    if not url_param_location:
        raise ValueError("Event must specify 'url' parameter location.")

    def observe_span_attrs(
        args: Mapping[str, object], result: object
    ) -> OtelAttributes:
        url = args[url_param_location]

        # This is a bit rough, but it's probably a good assumption that any url-like
        # argument is str-convertable. Known cases currently include:
        # - yarl.URL (aiohttp)
        # - httpx.URL (httpx)
        url = str(url)

        return _build_attrs(url)

    return SpanType.OutboundServiceCall, observe_span_attrs, None


def http_client_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    def http_client_attrs(args: Mapping[str, object], result: object) -> OtelAttributes:
        connection_obj: http.client.HTTPConnection = args["self"]
        incomplete_url: str = args["url"]

        scheme = (
            "https"
            if isinstance(connection_obj, http.client.HTTPSConnection)
            else "http"
        )
        host = connection_obj.host
        port = connection_obj.port
        full_url = f"{scheme}://{host}:{port}{incomplete_url}"

        return _build_attrs(full_url)

    return SpanType.OutboundServiceCall, http_client_attrs, None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode

quart_sources = [
    CompositeNode(
        {
            "module": "quart.wrappers.request",
            "class_name": "Request",
        },
        [
            {
                "method_name": ["base_url", "host_url", "url", "url_root"],
                "node_type": "URI",
                "tags": ["CROSS_SITE"],
            },
            {
                "method_name": "get_data",
                "target": "RETURN",
                "node_type": "BODY",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    {
        "module": "quart.wrappers.request",
        "class_name": "Body",
        "method_name": "__anext__",
        "node_type": "BODY",
        "tags": ["CROSS_SITE"],
    },
]


register_source_nodes(quart_sources)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

from typing import Callable
from collections.abc import Mapping
from collections.abc import Generator

from contrast_fireball import OtelAttributes, SpanType
from contrast_vendor import structlog as logging

from contrast.agent import scope
from contrast.agent.request_context import RequestContext
from contrast.policy_v2 import EventDict


EventHandler = Callable[..., Generator]
"""
A v2 policy event handler. The resulting generator must yield exactly once, and a
`result` of the original function call will be sent to the generator via this yield.
For example:

```python
def my_event_handler(instance, args, kwargs) -> Generator:
    # do pre-call work here
    result = yield
    # post-call work here
```
"""

EventHandlerBuilder = Callable[[EventDict], EventHandler]
"""
Builder function for a v2 policy event handler.
"""

AttrUpdateBuilder = Callable[
    [EventDict],
    tuple[
        SpanType,
        Callable[[Mapping[str, object], object], OtelAttributes],  # attribute builder
        "None | Callable[[Mapping[str, object]], str | None]",  # skip check function
    ],
]

logger = logging.getLogger("contrast")


def observe_handler_builder(
    attr_updater: AttrUpdateBuilder,
    event_dict: EventDict,
) -> EventHandler:
    """
    Builds an event handler that creates an observability child span around a function call.

    If the function is called within an existing observe scope, it will not create a new
    span, but will instead yield immediately.

    The builder needs an attribute updater that declares the type of action span to create,
    and how to generate attributes for that span based on the function arguments and result.
    The attribute updater may also return a skip check function that can be used to
    determine whether to skip the observe handler entirely based on the function arguments.
    """
    from contrast.agent import agent_state

    reporter = agent_state.module.reporting_client
    assert reporter is not None

    action_type, get_action_attributes, skip_check = attr_updater(event_dict)

    def observe_handler(
        context: RequestContext | None, args: Mapping[str, object]
    ) -> Generator:
        if context is None or scope.in_observe_scope():
            yield
            return
        with scope.observe_scope():
            if (trace := context.observability_trace) is None:
                yield
                return
            if skip_check and (reason := skip_check(args)):
                logger.debug(
                    "skipping observe handler",
                    reason=reason,
                    action_type=action_type,
                    args=args,
                )
                yield
                return

            with trace.child_span(action_type) as child_span:
                if child_span is None:
                    yield
                    return
                logger.debug(
                    "entered new child span",
                    child_span=child_span,
                    action_type=action_type,
                )
                result = yield
                action_attrs = get_action_attributes(args, result)
                child_span.update(action_attrs)
                logger.debug(
                    "updated child_span attributes",
                    child_span=child_span,
                    action_attrs=action_attrs,
                )

    return observe_handler



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import re
from contrast.agent.policy.patch_location_policy import PatchLocationPolicy
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.agent_lib import input_tracing
from contrast.agent.protect.rule.mode import Mode
from contrast.api.attack import Attack
from contrast.api.sample import Sample
from contrast.utils.loggers.logger import security_log_attack


class SqlInjection(BaseRule):
    """
    SQL Injection Protection rule
    """

    RULE_NAME = "sql-injection"

    def build_attack_with_match(
        self,
        candidate_string: str | None,
        evaluation: input_tracing.InputAnalysisResult | None = None,
        attack: Attack | None = None,
        **kwargs,
    ) -> Attack | None:
        assert evaluation is not None
        assert candidate_string is not None
        for match in re.finditer(
            re.compile(re.escape(evaluation.input.value)), candidate_string
        ):
            input_len = match.end() - match.start()
            agent_lib_evaluation = input_tracing.check_sql_injection_query(
                match.start(),
                input_len,
                input_tracing.DBType.from_str(kwargs["database"]),
                candidate_string,
            )
            if not agent_lib_evaluation:
                continue

            evaluation.attack_count += 1

            kwargs["start_idx"] = match.start()
            kwargs["end_idx"] = match.end()
            kwargs["boundary_overrun_idx"] = agent_lib_evaluation.boundary_overrun_index
            kwargs["input_boundary_idx"] = agent_lib_evaluation.input_boundary_index
            attack = self.build_or_append_attack(
                evaluation, attack, candidate_string, **kwargs
            )

        if attack is not None:
            attack.response = self.response_from_mode(self.mode)
            security_log_attack(attack, evaluation)

        return attack

    def build_attack_without_match(
        self,
        evaluation: input_tracing.InputAnalysisResult | None = None,
        attack: Attack | None = None,
        **kwargs,
    ) -> Attack | None:
        if self.mode == Mode.BLOCK_AT_PERIMETER:
            return super().build_attack_without_match(evaluation, attack, **kwargs)
        assert evaluation is not None
        if evaluation.score < 10:
            return None

        return next(
            (
                super(SqlInjection, self).build_attack_without_match(
                    full_eval, attack, **kwargs
                )
                for full_eval in evaluation.fully_evaluate()
                if full_eval.score >= 90
            ),
            None,
        )

    def build_sample(
        self,
        evaluation: input_tracing.InputAnalysisResult | None,
        candidate_string: str | None,
        **kwargs,
    ) -> Sample:
        assert evaluation is not None
        query = candidate_string
        sample = self.build_base_sample(evaluation)
        if query is not None:
            sample.details["query"] = query

        if "start_idx" in kwargs:
            sample.details["start"] = int(kwargs["start_idx"])

        if "end_idx" in kwargs:
            sample.details["end"] = int(kwargs["end_idx"])

        if "boundary_overrun_idx" in kwargs:
            sample.details["boundaryOverrunIndex"] = int(kwargs["boundary_overrun_idx"])

        if "input_boundary_idx" in kwargs:
            sample.details["inputBoundaryIndex"] = int(kwargs["input_boundary_idx"])

        return sample

    def infilter_kwargs(self, user_input: str, patch_policy: PatchLocationPolicy):
        return dict(database=patch_policy.module)

    def skip_protect_analysis(
        self, user_input: object, args: tuple, kwargs: dict[str, object]
    ) -> bool:
        """
        Some sql libraries use special objects (see from sqlalchemy import text)
        so we cannot just check if user_input is falsy.
        """
        return user_input is None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re


class PatternBuilder:
    QUOTE_PATTERN = "('|\")"
    MATCHED_PARENS = r"\(.*\)"
    STRING_CONCATENATE = (
        r'(\'|")(?:\s*)(?:\/\*.*\*\/)?(?:\s*)\+(?:\s*)(?:\/\*.*\*\/)?(?:\s*)(\'|")'
    )
    BLOCK_COMMENTS_AND_SPACES = r"(?:\s*)(?:\/\*.*\*\/)?(?:\s*)"
    JAVASCRIPT_IDENTIFIER = r"(?:[^\d({-][\w]*)"
    ESCAPE_REGEX = re.compile(r"([\[\]^()+*{}|?.-])")

    def __init__(self):
        self.pattern = []

    def build(self):
        # multiline = make anchors look for newline
        # dotall = make dot ('.') match newline
        return re.compile(
            "".join(self.pattern),
            flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL,
        )

    def quote(self):
        self.pattern.append(PatternBuilder.QUOTE_PATTERN)

        return self

    def some_or_no_block_comments_and_spaces(self):
        self.pattern.append(PatternBuilder.BLOCK_COMMENTS_AND_SPACES)

        return self

    def javascript_variable_identifiers(self):
        self.pattern.append(PatternBuilder.JAVASCRIPT_IDENTIFIER)

        return self

    def literally(self, string):
        self.pattern.append(self.escape(string))

        return self

    def anything(self):
        self.pattern.append(r".*")

        return self

    def any_of_these(self, string_list, escape=True, minimum=1, maximum=1):
        if not string_list or len(string_list) == 0:
            return self

        self.pattern.append("(")

        self.add_string_literals_to(
            string_list, "|", lambda x: self.escape(x) if escape else x
        )

        if minimum == maximum and minimum == 1:
            self.pattern.append(")")
        else:
            self.pattern.append("){" + str(minimum) + "," + str(maximum) + "}")

        return self

    def letters_at_least(self, length):
        self.pattern.append("[A-Za-z]{" + str(length) + ",}")

        return self

    def escape(self, string):
        # This also works; have not determined which is better
        # re.escape(string)

        return re.sub(PatternBuilder.ESCAPE_REGEX, lambda x: "\\" + x.group(0), string)

    def add_string_literals_to(self, source, separator, block):
        if not source:
            return

        self.pattern.append(block(source[0]))

        for string in source[1:]:
            if separator:
                self.pattern.append(separator)

            self.pattern.append(block(string))



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.configuration.agent_config import AgentConfig
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class DisableReaction:
    NAME = "DISABLE"
    ENABLE = "enable"
    MESSAGE = "Contrast received instructions to disable itself - Disabling now"

    @staticmethod
    def run(config: AgentConfig):
        logger.warning(DisableReaction.MESSAGE)

        if config:
            enable_option = config.get_option("enable")
            # We need to set the override here in case the Agent is set to enable by a higher precedence option like
            # env var.
            enable_option.override_value = False
            # But we also set the UI option b/c the value technically came from TeamServer
            enable_option.ui_value = False



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements AST rewriter for Contrast Agent
"""

from datetime import datetime, timezone
import importlib.abc
import importlib.machinery
import importlib.util
import sys
import ast
import copy
import builtins
import operator
import os
import time
from types import ModuleType

# Seems like it is necessary to import this prior to rewriting to avoid some weird partial import state
import tokenize  # noqa: F401

from contrast_vendor.wrapt.importer import _ImportHookChainedLoader

# NOTE: It is extremely important to limit the number of imports used by this
# module. It should be restricted to only those imports that are absolutely
# necessary for the operation of the rewriter, and ideally should include only
# built-in or standard library modules that are already imported by the
# interpreter prior to the evaluation of this module. It is *very* important
# that this module does *not* import the core `contrast` package since that
# would introduce a huge number of dependencies that we do not want.
# By limiting the number of dependencies for the rewriter module, we can ensure
# that a minimal number of modules are already imported prior to the
# application of the rewriter, which means we maximize the coverage of our
# rewriter in application and library code.

_CONTRAST_PACKAGES = ["contrast", "contrast_vendor", "contrast_rewriter"]
# disable `assert` in contrast modules. We don't want to do this for contrast_vendor,
# since some vendored packages might rely on `assert` behavior.
_PACKAGES_TO_OPTIMIZE = ["contrast"]


class _ContrastImportHookChainedLoader(_ImportHookChainedLoader):
    pass


class DeferredLogger:
    def __init__(self):
        self.messages = []

    def debug(self, message, *args, **kwargs):
        kwargs.setdefault(
            "time", datetime.now(tz=timezone.utc).isoformat().replace("+00:00", "Z")
        )
        self.messages.append(("debug", message, args, kwargs))


# Can't use our Namespace here since we don't want to import contrast package yet
class rewriter_module:
    logger = DeferredLogger()
    enabled: bool = False
    registry = set()


LOG_TIME_FORMAT = "%Y-%m-%d %H:%M:%S"
LOG_FORMAT = "{date_and_time} [{logger_name}] {msg}"


def log_stderr(msg: str, *, logger_name: str):
    """
    A basic logger that only depends on modules loaded before `sitecustomize`. Use
    sparingly, as these are highly visible and not configurable. We send all early logs
    to stderr (in structlog too) to minimize the probability of interfering with non-app
    process output in flex agent or operator environments.

    Use for log messages emitted before we have access to structlog.
    """
    date_and_time = time.strftime(LOG_TIME_FORMAT, time.localtime())
    output = LOG_FORMAT.format(
        date_and_time=date_and_time,
        logger_name=logger_name,
        msg=msg,
    )
    print(output, file=sys.stderr)


def _log(msg: str) -> None:
    log_stderr(msg, logger_name="contrast-rewriter")


def _load_module(source, module, filename, *, force_optimize=False):
    """
    Convenience method to compile and execute the given module source

    It seems like we do not need any exception handling here since any
    exception that occurs gets handled further up by the import machinery and
    causes it to fall back on the original loader. This is definitely good for
    us since it means that even if we mess up here somehow, it shouldn't
    prevent the original module from being loaded. It will just be loaded
    without our rewrites.

    If force_optimize is set, python optimizations equivalent to PYTHONOPTIMIZE=1 or -O
    will be applied to the new module (if not already applied globally).
    """
    optimize_flag = max(1, sys.flags.optimize) if force_optimize else sys.flags.optimize
    code = compile(source, filename, "exec", dont_inherit=True, optimize=optimize_flag)
    exec(code, module.__dict__)


def add_dependency_to_module(module: ModuleType, dependency: ModuleType):
    """
    Add a dependency to the provided module if it doesn't already exist.
    """
    add_dependency_to_globals(module.__dict__, dependency)


def add_dependency_to_globals(globs: dict, dependency: ModuleType):
    """
    Add a dependency to the provided globals if it doesn't already exist.
    """
    globs.setdefault(_contrast_module_name(dependency.__name__), dependency)


def _contrast_module_name(module_name: str):
    return f"_contrast__{module_name}"


def _get_top_level_module_name(fullname: str) -> str:
    fullname_split = fullname.split(".")
    if len(fullname_split) == 0:
        return ""
    return fullname_split[0]


class ContrastMetaPathFinder(importlib.abc.MetaPathFinder):
    def __init__(self) -> None:
        super().__init__()
        self.in_progress = set()

    def find_spec(self, fullname, path, target=None):
        """
        The finder is in charge of finding a module's "spec". The spec includes import
        machinery metadata about the module - including its name, source file path, and
        the loader, among others.

        Here, we first use importlib's utility to get the spec for the module
        about to be imported. The problem with this spec is that it also uses the
        default loader, which isn't what we want. To get around this, we reuse some
        metadata and generate a new spec that points at our loader.
        """
        if fullname in self.in_progress:
            # Here, we're within our own call to importlib.util.find_spec.
            # return None to allow the following MetaPathFinders to find the spec.
            return None
        self.in_progress.add(fullname)
        try:
            spec = importlib.util.find_spec(fullname, path)

            if (
                spec is None
                or spec.origin is None
                or not isinstance(spec.loader, importlib.machinery.SourceFileLoader)
            ):
                rewriter_module.logger.debug(
                    "Skipping non-source module",
                    module_name=fullname,
                    path=getattr(spec, "origin", "<unknown>"),
                )
                return None

            new_spec = importlib.util.spec_from_file_location(
                fullname,
                spec.origin,
                loader=ContrastRewriteLoader(fullname, spec.origin),
                submodule_search_locations=spec.submodule_search_locations,
            )
            if new_spec is None:
                return None

            loader = getattr(new_spec, "loader", None)

            if loader and not isinstance(loader, _ContrastImportHookChainedLoader):
                new_spec.loader = _ContrastImportHookChainedLoader(loader)

            rewriter_module.logger.debug(
                "Updated spec for module: fullname=%s, path=%s",
                fullname,
                spec.origin,
            )
            return new_spec
        finally:
            self.in_progress.discard(fullname)


class ContrastRewriteLoader(importlib.machinery.SourceFileLoader):
    def exec_module(self, module) -> None:
        """
        This method is responsible for actually doing the module `exec`-ing. We take
        control of this system and do the following:
        - read the original source file. We require pyc caching to be disabled for this
        - parse the source file into an AST
        - rewrite the AST
        - compile the AST into a code object
        - exec the code object

        Note that we add our custom add function to the module's globals. This prevents
        the need for import rewriting entirely.

        Contrast modules are not rewritten. Instead, we compile them with python
        optimizations by default (as if using PYTHONOPTIMIZE=1 or -O). This removes
        `assert` and `if __debug__:` statements from contrast production code.
        """
        original_source_code = None
        filename = self.path

        # May be None in some cases such as for namespace packages
        if filename is None:
            return

        rewriter_module.logger.debug(
            "ContrastRewriteLoader started executing module",
            filename=filename,
            module_name=self.name,
        )

        try:
            original_source_code = self.get_source(self.name)
            tree = ast.parse(original_source_code)
        except Exception as ex:
            rewriter_module.logger.debug(
                "WARNING: failed to parse module AST",
                filename=filename,
                exc_info=ex,
                module_name=self.name,
            )

            _load_module(original_source_code, module, filename)
            return

        if _get_top_level_module_name(self.name) in _PACKAGES_TO_OPTIMIZE:
            _load_module(
                original_source_code,
                module,
                filename,
                force_optimize=(not os.environ.get("CONTRAST_TESTING")),
            )
            return

        if module_was_rewritten(self.name):
            rewriter_module.logger.debug(
                "WARNING: module appears to have been already rewritten; will not rewrite again",
                filename=filename,
            )
        else:
            try:
                propagation_rewriter = PropagationRewriter()
                propagation_rewriter.visit(tree)
                propagation_rewriter.populate_dependencies(module)
            except Exception as ex:
                rewriter_module.logger.debug(
                    "WARNING: failed to rewrite module",
                    filename=filename,
                    exc_info=ex,
                    module_name=self.name,
                )

        rewriter_module.registry.add(self.name)

        _load_module(tree, module, filename)
        rewriter_module.logger.debug(
            "ContrastRewriteLoader finished executing module",
            filename=filename,
            module_name=self.name,
        )


class PropagationRewriter(ast.NodeTransformer):
    # all_possible_injected_modules holds a set of all modules
    # that could be injected into rewritten ASTs by any instance.
    #
    # This is a broad set. See the injected_modules instance
    # attribute for only the modules injected by one instance.
    all_possible_injected_modules = {builtins, operator}

    def __init__(self) -> None:
        super().__init__()
        self.injected_modules = set()

    def populate_dependencies(self, module: ModuleType):
        """
        Adds dependencies that were required by past rewrites.
        """
        for dependency in self.injected_modules:
            add_dependency_to_module(module, dependency)

    def _copy_with_context(self, node, context):
        node = copy.copy(node)
        node.ctx = context
        return node

    def _make_attr(self, module: ModuleType, attr: str):
        self.injected_modules.add(module)
        return ast.Attribute(
            value=ast.Name(id=_contrast_module_name(module.__name__), ctx=ast.Load()),
            attr=attr,
            ctx=ast.Load(),
        )

    def visit_BinOp(self, binop: ast.BinOp):
        """
        If we see an "Add" or a "Mod" binary operation, replace it with a call to our custom add/modulo
        function, which includes all necessary instrumentation.
        """
        binop.left = self.visit(binop.left)
        binop.right = self.visit(binop.right)

        if isinstance(binop.op, ast.Mod):
            binop_replacement = ast.Call(
                func=self._make_attr(operator, "mod"),
                args=[binop.left, binop.right],
                keywords=[],
            )
            ast.copy_location(binop_replacement, binop)
            return ast.fix_missing_locations(binop_replacement)

        if not isinstance(binop.op, ast.Add):
            return binop

        binop_replacement = ast.Call(
            func=self._make_attr(operator, "add"),
            args=[binop.left, binop.right],
            keywords=[],
        )
        ast.copy_location(binop_replacement, binop)
        return ast.fix_missing_locations(binop_replacement)

    def visit_AugAssign(self, node: ast.AugAssign):
        """
        If we see an "Append", `+=` operation, rewrite it as a `+`.
        """
        node.value = self.visit(node.value)

        if not isinstance(node.op, ast.Add):
            return node

        target = left = None
        if isinstance(node.target, ast.Name):
            name = ast.Name(id=node.target.id)
            target = self._copy_with_context(name, ast.Store())
            left = self._copy_with_context(name, ast.Load())
        else:
            target = node.target
            left = self._copy_with_context(target, ast.Load())

        call_contrast_append_node = ast.Assign(
            targets=[target],
            value=ast.Call(
                func=self._make_attr(operator, "iadd"),
                args=[self.visit(left), node.value],
                keywords=[],
            ),
        )
        ast.copy_location(call_contrast_append_node, node)
        return ast.fix_missing_locations(call_contrast_append_node)

    def visit_JoinedStr(self, node: ast.JoinedStr):
        node.values = [self.visit(value) for value in node.values]
        call_node = ast.Call(
            func=ast.Attribute(value=ast.Constant(""), attr="join", ctx=ast.Load()),
            args=[ast.List(elts=node.values, ctx=ast.Load())],
            keywords=[],
        )
        ast.copy_location(call_node, node)
        return ast.fix_missing_locations(call_node)

    def visit_FormattedValue(self, node: ast.FormattedValue):
        node.value = self.visit(node.value)

        # See https://docs.python.org/3/library/ast.html#ast.FormattedValue
        # for the conversion codes.
        if node.conversion != -1:
            if node.conversion == 115:
                conversion = "str"
            elif node.conversion == 114:
                conversion = "repr"
            elif node.conversion == 97:
                conversion = "ascii"
            else:
                rewriter_module.logger.debug(
                    "unexpected conversion code", code=node.conversion
                )
                return node

            node.value = ast.Call(
                func=self._make_attr(builtins, conversion),
                args=[node.value],
                keywords=[],
            )

        call_node = ast.Call(
            func=self._make_attr(builtins, "format"),
            args=[node.value, node.format_spec] if node.format_spec else [node.value],
            keywords=[],
        )
        ast.copy_location(call_node, node)
        return ast.fix_missing_locations(call_node)


def _non_contrast_module_filter(module_item) -> bool:
    return (
        _get_top_level_module_name(getattr(module_item[1], "__package__", "") or "")
        not in _CONTRAST_PACKAGES
    )


def _log_imported_modules() -> None:
    module_map = {
        importlib.machinery.BuiltinImporter: "builtin",
        importlib.machinery.FrozenImporter: "frozen",
    }

    all_modules = list(filter(_non_contrast_module_filter, sys.modules.items()))

    rewriter_module.logger.debug(
        "the following %d modules are already imported", len(all_modules)
    )
    for name, module in sorted(all_modules):
        module_type = module_map.get(getattr(module, "__loader__", None), "source")
        rewriter_module.logger.debug("%-20s type=%s", name, module_type)


def _hook_assertion_rewrites(rewrite_module):
    from contrast_vendor.wrapt import function_wrapper

    # This hook enables us to apply our rewriter before assertion rewrites are applied
    def rewrite_asserts(wrapped, _, args, kwargs):
        file_ast = args[0]
        PropagationRewriter().visit(file_ast)
        wrapped(*args, **kwargs)

    # This hook ensures that we add our contrast-specfic functions to the rewritten module
    def exec_module(wrapped, _, args, kwargs):
        module = args[0]
        for dep in PropagationRewriter.all_possible_injected_modules:
            add_dependency_to_module(module, dep)
        wrapped(*args, **kwargs)

    rewrite_module.rewrite_asserts = function_wrapper(rewrite_asserts)(
        rewrite_module.rewrite_asserts
    )
    rewrite_module.AssertionRewritingHook.exec_module = function_wrapper(exec_module)(
        rewrite_module.AssertionRewritingHook.exec_module
    )


def register_assertion_rewrite_hooks():
    """
    Register hooks for pytest's assertion rewriter

    This is only to be used for internal testing purposes. It enables our
    rewrites to be compatible with pytest's assertion rewrites.
    """
    from contrast_vendor.wrapt import register_post_import_hook

    register_post_import_hook(_hook_assertion_rewrites, "_pytest.assertion.rewrite")


def register():
    """
    Register our rewriter with the import system. After this call, any newly imported
    modules (from source code) will use our custom rewriter.

    Note that because this function is defined in the same module that defines our add
    replacement function, we never have to worry about rewriting the addition in the
    replacement function itself. If that were to occur, we would get an infinite
    recursion.
    """
    if is_rewriter_enabled():
        rewriter_module.logger.debug("Rewriter already enabled, not applying again")
        return

    # Useful for debugging, but slow
    # _log_imported_modules()

    sys.meta_path.insert(0, ContrastMetaPathFinder())

    rewriter_module.logger.debug("enabled AST rewriter")

    rewriter_module.enabled = True


def deregister():
    """
    Remove our rewriter from the import system. Modules that were loaded by our rewriter
    will remain rewritten.

    Return True if we find and deregister our machinery, False otherwise.
    """
    for i, finder in enumerate(sys.meta_path.copy()):
        if isinstance(finder, ContrastMetaPathFinder):
            sys.meta_path.pop(i)
            rewriter_module.enabled = False
            return True
    return False


def initialize_rewriter_logger(agent_logger):
    if not isinstance(rewriter_module.logger, DeferredLogger):
        agent_logger.debug("WARNING: cannot process deferred rewriter logs")
        return

    if len(rewriter_module.logger.messages) > 0:
        agent_logger.debug("---- Beginning of deferred rewriter logs ----")
        for level, message, args, kwargs in rewriter_module.logger.messages:
            getattr(agent_logger, level)(message, *args, **kwargs)
        agent_logger.debug("---- End of deferred rewriter logs ----")
        rewriter_module.logger.messages.clear()

    agent_logger.debug("Setting rewriter logger to agent logger")
    rewriter_module.logger = agent_logger


def is_rewriter_enabled() -> bool:
    return rewriter_module.enabled


def module_was_rewritten(name: str) -> bool:
    return name in rewriter_module.registry


def clear_registry():
    rewriter_module.registry.clear()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import builtins
import io
import inspect
import os
import sys


from contrast.agent.assess.utils import get_properties
from contrast.agent.policy import patch_manager
from contrast.agent.policy.applicator import (
    apply_module_patches,
    reverse_module_patches,
)
from contrast.agent.scope import pop_contrast_scope
from contrast.assess_extensions import smart_setattr
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    repatch_module,
    register_module_patcher,
    unregister_module_patcher,
)


def property_getter(self):
    """This property makes pathlib.Path instances look like tracked strings"""
    # Getting the string representation could cause propagation to occur
    with pop_contrast_scope():
        string = str(self)

    return get_properties(string)


def create_adjusted_patch(patch):
    def adjusted_patch(path, *args, **kwargs):
        # We cast the path to str only for compatiblity with PY35. In all later
        # versions of Python, the underlying system functions accept pathlib.Path
        # objects as parameters, and so the cast is redundant. Strictly speaking, this
        # cast means that we don't need the cs__property attribute for any cases, but
        # it is good to have for future-proofing.
        return patch(str(path), *args, **kwargs)

    return staticmethod(adjusted_patch)


def is_wrapped_staticmethod(obj):
    return isinstance(obj, staticmethod) and hasattr(obj.__func__, "__wrapped__")


@fail_quietly("Failed to apply pathlib patches")
def patch_pathlib(pathlib_module):
    """
    Apply patches to the pathlib module

    We abuse the patch manager here to actually repatch a _class_.  This is necessary
    because pathlib stores references to a bunch of os module functions within an
    accessor class. Since pathlib is part of the standard library, this happens
    before our patching machinery is applied, which means the accessor retains
    references to the _original_ functions, and not our patched versions. We can't
    simply run repatch_module on pathlib itself because the references are in the
    class definition and not at module level. But if we explicitly repatch the
    _class_, everything works out just fine.

    Currently we would potentially miss cases where a custom accessor class was used
    (assuming the custom accessor was also defined before our patches applied). In
    theory, we could handle this by repatching the given accessor at the time each
    pathlib.Path is instantiated. However, given the fact that this functionality isn't
    even documented, I don't think it's worth pursuing right now.

    """
    # In order for pathlib patches to work, the os and io patches *must* be applied first
    # The best way to ensure that this occurs in the right order is by applying them
    # explicitly here before doing anything else.
    apply_module_patches(builtins)
    apply_module_patches(io)
    # Make sure builtin patches are applied to io module
    repatch_module(io)

    apply_module_patches(os)

    repatch_module(pathlib_module)

    # Using this property enables pathlib.Path instances to look like tracked strings
    # for the purposes of triggering vulnerabilities.
    pathprop = property(fget=property_getter)
    smart_setattr(pathlib_module.Path, "cs__properties", pathprop)

    # Overly-cautious safety check
    accessor_cls = getattr(pathlib_module, "_NormalAccessor", None)
    if accessor_cls is not None:
        for attr_name, attr in accessor_cls.__dict__.items():
            if not inspect.isroutine(attr):
                continue

            patch = patch_manager.get_patch(attr)

            if patch is None:
                continue

            # This does an end-run around the patch manager, but since it's a very
            # special case, we shouldn't be too concerned about reverse patching.
            smart_setattr(accessor_cls, attr_name, create_adjusted_patch(patch))

    # This is probably not necessary in most real-world environments but it
    # ensures that our instrumented `join` method is used in cases where the
    # flavours are initialized prior to our patch application.
    for name in ["_posix_flavour", "_windows_flavour"]:
        flavour = getattr(pathlib_module, name, None)
        if flavour is not None:
            smart_setattr(flavour, "join", flavour.sep.join)


def register_patches():
    register_module_patcher(patch_pathlib, "pathlib")


def reverse_patches():
    reverse_module_patches(os)
    reverse_module_patches(io)
    reverse_module_patches(builtins)

    unregister_module_patcher("pathlib")
    pathlib_module = sys.modules.get("pathlib")
    if not pathlib_module:
        return

    if hasattr(pathlib_module, "_NormalAccessor"):
        patch_manager.reverse_patches_by_owner(pathlib_module._NormalAccessor)

    delattr(pathlib_module.Path, "cs__properties")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode


_CYTHON_REQUEST_POLICY = [
    {
        "method_name": [
            "uri",
            "url",
            "relative_uri",
            "prefix",
        ],
        "node_type": "URI",
        "tags": ["CROSS_SITE"],
    },
    {
        "method_name": [
            "forwarded_uri",
            "forwarded_prefix",
        ],
        "node_type": "HEADER",
        "tags": ["NO_NEWLINES"],
    },
]
"""
Cythonized Falcon cannot be rewritten, so we
lose dataflow propagation through string concats.
We need policy to cover this gap in the following
properties on falcon.request.Request
"""


falcon_sources = [
    CompositeNode(
        {
            "module": "falcon.request",
            "class_name": "Request",
        },
        [
            {
                "method_name": ["get_param", "params"],
                "node_type": "QUERYSTRING",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
            {
                "method_name": "get_media",
                "node_type": "BODY",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
        ]
        + _CYTHON_REQUEST_POLICY,
    ),
    CompositeNode(
        {
            "module": "falcon.asgi.request",
            "class_name": "Request",
        },
        [
            {
                "method_name": "get_param",
                "node_type": "QUERYSTRING",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
            {
                "method_name": "get_media",
                "node_type": "BODY",
                "tags": ["NO_NEWLINES", "CROSS_SITE"],
            },
        ]
        + _CYTHON_REQUEST_POLICY,
    ),
    {
        "module": "falcon.asgi.stream",
        "class_name": "BoundedStream",
        "method_name": "read",
        "node_type": "BODY",
        "tags": ["NO_NEWLINES", "CROSS_SITE"],
    },
    {
        "module": "falcon.asgi.reader",
        "class_name": "BufferedReader",
        "method_name": "read",
        "node_type": "BODY",
        "tags": ["NO_NEWLINES", "CROSS_SITE"],
    },
]


register_source_nodes(falcon_sources)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
import re


from enum import Enum, auto

from contrast.utils.decorators import fail_loudly


from contrast_fireball import AssessEventAction, AssessFinding
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class InputExclusionSourceType(Enum):
    COOKIE = auto()
    PARAMETER = auto()
    HEADER = auto()
    BODY = auto()
    QUERYSTRING = auto()
    UNKNOWN = auto()


@fail_loudly("Unable to compile regex", log_level="warning")
def safe_re_compile(pattern, flags=0):
    return re.compile(pattern, flags=flags)


def strs_to_regexes(patterns: list[str], ignorecase=0):
    """
    Safely compiles a list of pattern strings into regex patterns.
    If a string cannot be compiled, it is skipped.
    """
    return [
        compiled_regex
        for pattern in patterns
        if (compiled_regex := safe_re_compile(pattern, flags=ignorecase))
    ]


class Exclusions:
    """
    Container for all exclusions
    """

    @staticmethod
    def init_input_exclusions_container():
        input_exclusions = dict(
            HEADER=[], HEADER_KEY=[], PARAMETER=[], COOKIE=[], BODY=[], QUERYSTRING=[]
        )

        # Aliases for the same source type
        input_exclusions["HEADER"] = input_exclusions["HEADER_KEY"]
        input_exclusions["MULTIPART_FORM_DATA"] = input_exclusions["BODY"]
        input_exclusions["MULTIPART_CONTENT_DATA"] = input_exclusions["BODY"]

        return input_exclusions

    def __init__(self, config: Mapping):
        active_modes = set()
        if config["assess.enable"]:
            active_modes.add("assess")
        if config["protect.enable"]:
            active_modes.add("defend")

        def active(exclusions):
            return (
                excl
                for excl in exclusions
                if set(excl.get("modes", {})).intersection(active_modes)
            )

        self.url_exclusions = [
            UrlExclusion(excl)
            for excl in active(config.get("application.url_exclusions", []))
        ]

        has_named_exclusion = False
        input_exclusions = self.init_input_exclusions_container()
        for exc in active(config.get("application.input_exclusions", [])):
            input_exclusions[exc.get("input_type")].append(InputExclusion(exc))
            has_named_exclusion = True
        self.input_exclusions = input_exclusions if has_named_exclusion else None

    def evaluate_assess_trigger_time_exclusions(self, context, finding):
        # returns True if we do not report finding
        return any(
            exc.match_in_finding(finding)
            for exc in context.input_exclusions_trigger_time
        )

    def evaluate_input_exclusions(
        self, context, source_type: str, source_name: str, mode=None
    ):
        # Evaluate all exclusions against the current source
        if context.input_exclusions is None:
            return False

        exclusions = context.input_exclusions.get(source_type, None)
        if exclusions is None:
            return False

        for exc in exclusions:
            if mode and mode not in exc.modes:
                continue

            if exc.match(
                context,
                source_type=source_type,
                source_name=source_name,
            ):
                logger.debug(
                    "The input exclusion rule named '%s' matched on the input name '%s' for the input type of '%s'",
                    exc.exclusion_name,
                    source_name,
                    exc.input_type,
                )
                return True

        return False

    def evaluate_input_exclusions_url(
        self, context, source_type: str, path: str, mode=None
    ):
        # Evaluate all exclusions against the current source
        if context.input_exclusions is None:
            return False

        exclusions = context.input_exclusions.get(source_type, None)
        if exclusions is None:
            return False

        for exc in exclusions:
            if mode and mode not in exc.modes:
                continue

            if exc.match_type == "ALL":
                logger.debug(
                    "The input url exclusion rule named '%s' matched on the path '%s' for the input type of '%s'",
                    exc.exclusion_name,
                    path,
                    exc.input_type,
                )
                return True

            for url_regex in exc.url_regexes:
                if url_regex.search(path):
                    logger.debug(
                        "The input url exclusion rule named '%s' matched on the path '%s' for the input type of '%s'",
                        exc.exclusion_name,
                        path,
                        exc.input_type,
                    )
                    return True

        return False

    def evaluate_url_exclusions(self, context, path: str) -> None:
        """
        This function evaluates all exclusions depending on the request URL and updates the request context to contain
        the list of disabled assess and protect rules to be evaluated at trigger time for url exclusions
        """
        for exc in self.url_exclusions:
            exc.evaluate(context, path)

    def set_input_exclusions_by_url(self, context, path: str):
        """
        Evaluates the set of input exclusions that apply to this path. Update request
        context with input exclusions to apply.
        """
        if self.input_exclusions is None:
            return

        context.input_exclusions = self.init_input_exclusions_container()
        context.input_exclusions_trigger_time = []

        has_match = False

        for input_type, exclusions in self.input_exclusions.items():
            for exc in exclusions:
                if exc.url_regexes and not any(
                    pattern.fullmatch(path) for pattern in exc.url_regexes
                ):
                    continue

                if exc.protect_rules or exc.assess_rules:
                    context.input_exclusions_trigger_time.append(exc)
                else:
                    context.input_exclusions[input_type].append(exc)

                has_match = True

        if not has_match:
            # No exclusions for this request
            context.input_exclusions = None


class BaseExclusion:
    def __init__(self, exclusion: Mapping):
        self.exclusion_name = exclusion.get("name")
        self.url_regexes = strs_to_regexes(exclusion.get("urls", []))

        self.protect_rules = exclusion.get("protect_rules", [])
        self.assess_rules = exclusion.get("assess_rules", [])

        self.match_type = exclusion.get("match_strategy")
        self.modes = exclusion.get("modes", [])


class InputExclusion(BaseExclusion):
    def __init__(self, exclusion: Mapping):
        super().__init__(exclusion)

        self.input_name_regex = None

        self.input_type: InputExclusionSourceType = getattr(
            InputExclusionSourceType,
            exclusion.get("input_type", ""),
            InputExclusionSourceType.UNKNOWN,
        )

        if self.input_type == InputExclusionSourceType.UNKNOWN:
            logger.error(
                "Invalid input exclusion type for the exclusion", exclusion=exclusion
            )

        if input_name := exclusion.get("input_name"):
            # Adding a check for type cookie to ignore case as per requirements
            # https://github.com/Contrast-Security-Inc/platform-specifications/blob/main/exclusions/EXCLUSIONS_INPUT.md
            flags = (
                re.IGNORECASE
                if self.input_type == InputExclusionSourceType.COOKIE
                else 0
            )
            self.input_name_regex = safe_re_compile(input_name, flags)

    def is_body_input_type(self, source_type: str):
        return self.input_type == InputExclusionSourceType.BODY or source_type in [
            "BODY",
            "MULTIPART_FORM_DATA",
            "MULTIPART_CONTENT_DATA",
        ]

    def is_querystring_input_type(self, source_type: str):
        return (
            self.input_type == InputExclusionSourceType.QUERYSTRING
            or source_type == "QUERYSTRING"
        )

    def match_in_finding(self, finding: AssessFinding):
        # pylint: disable=too-many-nested-blocks
        if finding.rule_id in self.assess_rules:
            for event in finding.events:
                if event.action == AssessEventAction.CREATION:
                    for trace_event in event.event_sources:
                        event_type = trace_event.source_type.name
                        event_src_name = trace_event.source_name

                        if event_type == self.input_type.name:
                            exclude = False

                            if (
                                self.input_name_regex is not None
                                and self.input_name_regex.fullmatch(event_src_name)
                            ):
                                exclude = True

                            if self.input_type in [
                                InputExclusionSourceType.QUERYSTRING,
                                InputExclusionSourceType.BODY,
                            ]:
                                exclude = True

                            if exclude:
                                logger.debug(
                                    "The input exclusion rule named '%s' matched on the input name '%s' for the input type of '%s' for the rule '%s'",
                                    self.exclusion_name,
                                    event_src_name,
                                    event_type,
                                    finding.rule_id,
                                )
                                return exclude

        return False

    def match(
        self,
        context,
        source_type: str | None = None,
        source_name: str | None = None,
    ):
        if source_name is None and self.input_type not in [
            InputExclusionSourceType.QUERYSTRING,
            InputExclusionSourceType.BODY,
        ]:
            return False

        if self.is_body_input_type(source_type) or self.is_querystring_input_type(
            source_type
        ):
            return True

        return self.input_name_regex.fullmatch(source_name) is not None


class UrlExclusion(BaseExclusion):
    @fail_loudly("Unable to ignore request", return_value=False)
    def evaluate(self, context, path: str):
        """
        Determine if the given path exactly matches any of the
        configured urls for this exclusion rule. This function modifies the
        request context to notify rules if they should be disabled for this url.

        @param context: request context
        @param path: path for current request
        @return This function returns True if a match was found
        @rtype: bool
        """
        for pattern in self.url_regexes:
            if pattern.fullmatch(path):
                logger.debug("Path %s matched on pattern %s", path, pattern)
                self.update_disabled_rules(context)
                return True

        return False

    def update_disabled_rules(self, context):
        if "assess" in self.modes:
            if self.assess_rules:
                context.excluded_assess_rules.extend(self.assess_rules)
            else:
                context.assess_enabled = False

        if "defend" in self.modes:
            if self.protect_rules:
                context.excluded_protect_rules.extend(self.protect_rules)
            else:
                context.protect_enabled = False



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import importlib.abc
import importlib.util

import sys
from types import ModuleType

import contrast
from contrast.agent import scope
from contrast.agent.settings import Settings
from contrast.applies.assess import unsafe_code_execution
from contrast_vendor.wrapt.importer import _ImportHookChainedLoader


def register_path_finder():
    """
    Add the ContrastMetaPathFinder to the sys.meta_path if it is not already present.
    """
    if not any(
        isinstance(path_finder, ContrastMetaPathFinder) for path_finder in sys.meta_path
    ):
        sys.meta_path.insert(0, ContrastMetaPathFinder())


def remove_path_finder():
    """
    Remove the ContrastMetaPathFinder from the sys.meta_path if it is present.
    """
    for path_finder in sys.meta_path:
        if isinstance(path_finder, ContrastMetaPathFinder):
            sys.meta_path.remove(path_finder)
            break


class ContrastMetaPathFinder(importlib.abc.MetaPathFinder):
    def __init__(self) -> None:
        super().__init__()
        self.in_progress = {}

    def find_spec(self, fullname, path, target=None):
        if fullname in self.in_progress:
            return None
        self.in_progress[fullname] = True
        try:
            analyze_for_unsafe_code_execution(fullname)

            spec = importlib.util.find_spec(fullname)
            if spec is None:
                return None

            loader = getattr(spec, "loader", None)
            if loader and not isinstance(loader, _ContrastImportHookChainedLoader):
                spec.loader = _ContrastImportHookChainedLoader(loader)

            return spec
        finally:
            del self.in_progress[fullname]


def analyze_for_unsafe_code_execution(module_name: str):
    if scope.in_contrast_scope():
        return
    with scope.contrast_scope():
        context = contrast.REQUEST_CONTEXT.get()
        if context and context.assess_enabled:
            unsafe_code_execution.apply_rule(
                "importlib", "__import__", None, (module_name,), {}
            )


class _ContrastImportHookChainedLoader(_ImportHookChainedLoader):
    @scope.contrast_scope()
    @scope.observe_scope()  # don't report file-open-create for imported source files
    def _self_exec_module(self, module: ModuleType) -> None:
        in_scope = False
        try:
            with scope.pop_contrast_scope():
                in_scope = scope.in_contrast_scope()
                return super()._self_exec_module(module)
        finally:
            if (
                not in_scope
                and Settings().is_inventory_enabled()
                and (context := contrast.REQUEST_CONTEXT.get()) is not None
            ):
                context.observed_libraries.add_module(module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.mode import Mode


class MalformedHeader(BaseRule):
    """
    Malformed Header Protection rule
    """

    RULE_NAME = "malformed-header"

    @property
    def mode(self):
        """
        Always block at perimeter
        """
        return Mode.BLOCK_AT_PERIMETER



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


import base64
import contextlib
from dataclasses import dataclass
from http.cookiejar import CookieJar, DefaultCookiePolicy
import queue
import threading
from collections.abc import Generator
import uuid

from contrast.agent.request import Request
from contrast.configuration.agent_config import AgentConfig
from contrast.configuration.config_option import DEFAULT_VALUE_SRC
from contrast.utils import timer
import contrast_fireball
import contrast

from contrast.agent import scope
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.settings import Settings
from contrast.reporting.request_audit import RequestAudit
from contrast.reporting.teamserver_messages.base_ts_message import (
    PYTHON,
    b64url_stripped,
)
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast_vendor import structlog as logging

from .teamserver_messages import BaseTsMessage
from contrast.reporting import teamserver_messages

logger = logging.getLogger("contrast")

REPORTING_CLIENT_THREAD_NAME = "ContrastReportingClient"

MAX_ATTEMPTS = 2
ERROR_STATUS_CODE = -1


class ReportingClient(threading.Thread):
    def __init__(self, instance_id: str | None = None):
        super().__init__(name=REPORTING_CLIENT_THREAD_NAME, daemon=True)
        from contrast.agent import agent_state

        self.instance_id = instance_id or str(uuid.uuid4())
        self._stopper = object()  # sentinel sent over message_q to stop the thread
        self.message_q = queue.Queue(maxsize=128)
        self.settings = Settings()

        self.init_certs(self.settings)
        self.proxies = (
            self.settings.build_proxy_url() if self.settings.is_proxy_enabled else {}
        )
        # disable persisting cookies, per Architecture team recommendation.
        self.cookies = CookieJar(policy=DefaultCookiePolicy(allowed_domains=[]))
        server_name_b64 = b64url_stripped(agent_state.get_server_name())
        server_path_b64 = b64url_stripped(self.settings.get_server_path())
        server_type_b64 = b64url_stripped(self.settings.server_type)
        auth_header = f"{self.settings.api_user_name}:{self.settings.api_service_key}"
        self.always_headers = {
            # the Authorization header must not have its padding stripped
            "Authorization": base64.urlsafe_b64encode(auth_header.encode()).decode(),
            "API-Key": self.settings.api_key,
            "Server-Name": server_name_b64,
            "Server-Path": server_path_b64,
            "Server-Type": server_type_b64,
            "X-Contrast-Agent": f"{PYTHON} {contrast.__version__}",
            "X-Contrast-Header-Encoding": "base64",
            "X-Contrast-Reporting-Instance": self.instance_id,
            "Application-Language": PYTHON,
            "Application-Name": b64url_stripped(agent_state.get_app_name()),
            "Application-Path": b64url_stripped(self.settings.app_path),
        }

        self.request_audit = (
            RequestAudit(self.settings.config)
            if self.settings.config.is_request_audit_enabled
            else None
        )
        if self.request_audit:
            self.request_audit.prepare_dirs()

    @fail_quietly(return_value=False)
    def initialize_application(self, config: AgentConfig, server_type="") -> bool:
        """
        Initialize the application with the given configuration.

        Returns True if the application was successfully initialized.
        """
        attempt = 1
        while True:
            if _success := self._agent_startup():
                return True

            attempt += 1
            if attempt > MAX_ATTEMPTS:
                break

            logger.debug(
                "App initialization did not connect -  will retry sleeping for 1 second"
            )
            timer.sleep(1)

        return False

    def _agent_startup(self) -> bool:
        msg = teamserver_messages.AgentStartup()
        if (response := self.send_message(msg)) is not None:
            msg.process_response(response, self)
            if response.status_code < 500:
                return True
            elif response.status_code == 500:
                raise RuntimeError("Unexpected 500 response code from AgentStartup")
        return False

    def new_discovered_routes(self, routes: set[contrast_fireball.DiscoveredRoute]):
        self.add_message(teamserver_messages.ApplicationInventory(routes))

    def new_observed_route(self, route: contrast_fireball.ObservedRoute):
        self.add_message(teamserver_messages.ObservedRoute(route))

    def new_findings(
        self, findings: list[contrast_fireball.AssessFinding], request: Request | None
    ):
        self.add_message(teamserver_messages.Preflight(findings, request))

    def new_inventory_components(
        self, components: list[contrast_fireball.InventoryComponent]
    ):
        self.add_message(
            teamserver_messages.ApplicationActivity(inventory_components=components)
        )

    def new_libraries(self, libraries: list[contrast_fireball.Library]):
        self.add_message(teamserver_messages.ApplicationUpdate(libraries))

    def new_library_observations(
        self, observations: list[contrast_fireball.LibraryObservation]
    ):
        self.add_message(teamserver_messages.LibraryUsage(observations))

    @contextlib.contextmanager
    def observability_trace(
        self,
        *,
        send_trace: bool,
        attributes: contrast_fireball.OtelAttributes | None = None,
    ) -> Generator:
        logger.error(
            "Observe mode requires `api.reporting_client = fireball`, but direct"
            " reporting is in use. Disabling observe mode."
        )
        from contrast.agent import agent_state

        agent_state.set_observe_enabled({"observe.enable": False})
        yield None

    def init_certs(self, settings: Settings) -> None:
        self.verify = True
        self.cert = None

        certificate_enable_config_option = settings.config.get_option(
            "api.certificate.enable"
        )
        assert certificate_enable_config_option is not None
        if not certificate_enable_config_option.value():
            return

        # custom certificate settings
        if settings.ca_file:
            self.verify = settings.ca_file
        if settings.client_cert_file and settings.client_private_key:
            self.cert = (
                settings.client_cert_file,
                settings.client_private_key,
            )
        if settings.ignore_cert_errors:
            self.verify = False
            logger.warning("Certificate verification is disabled.")

        # misconfig checks
        if (settings.client_cert_file and not settings.client_private_key) or (
            not settings.client_cert_file and settings.client_private_key
        ):
            logger.error(
                "Unable to communicate with Contrast. "
                "Certificate configuration is not set properly. "
                "Certificate PEM file or private key PEM file is missing.",
                cert_file=settings.client_cert_file,
                key_file=settings.client_private_key,
            )
            DisableReaction.run(settings.config)
            return

        if certificate_enable_config_option.source() != DEFAULT_VALUE_SRC and not any(
            (settings.ca_file, settings.client_cert_file, settings.client_private_key)
        ):
            logger.warning(
                "Certificate configuration is explicitly enabled, but no certificate files are set."
            )

    def stop(self):
        self.message_q.put(self._stopper)
        self.join()

    def run(self):
        with scope.contrast_scope():
            logger.debug("Starting reporting thread")

            while self.settings.is_agent_config_enabled():
                try:
                    msg = self.message_q.get(block=True, timeout=5)
                    if msg is self._stopper:
                        break
                    response = self.send_message(msg)
                    msg.process_response(response, self)
                except queue.Empty:
                    pass
                except Exception as e:
                    logger.debug(
                        "WARNING: reporting client failed to send message", exc_info=e
                    )

    @fail_loudly("Failed to send message to Contrast")
    def send_message(self, msg: BaseTsMessage):
        """
        _send_message sends msg to Teamserver and returns the response.

        It is the caller's responsibility to handle any network exceptions.

        See send_message for the public interface, which handles exceptions.
        """
        status_code = ERROR_STATUS_CODE
        msg_name = msg.class_name

        url = msg.base_url + msg.path
        logger.debug("Sending %s message to Teamserver: %s", msg_name, url)
        response = msg.request_method(
            url,
            json=msg.body,
            headers={**self.always_headers, **msg.extra_headers},
            cookies=self.cookies,
            allow_redirects=False,
            proxies=self.proxies,
            verify=self.verify,
            cert=self.cert,
        )

        try:
            status_code = response.status_code
            msg_success_status = response.json().get("success")
            messages = response.json().get("messages")
            if not (msg_success_status or status_code == 200):
                logger.error(
                    "Failure on Contrast UI processing request reason - (%s): %s",
                    messages,
                    status_code,
                )
        except Exception as e:
            if status_code == ERROR_STATUS_CODE:
                logger.debug(
                    "Failed to receive response from Contrast UI: %s ",
                    e,
                )

        logger.debug("Contrast UI response (%s): %s", msg_name, status_code)

        if self.request_audit:
            self.request_audit.audit(msg, response)

        msg.sent()
        return response

    def add_message(self, msg):
        if msg is None or not isinstance(msg, BaseTsMessage):
            return

        logger.debug("Adding msg to reporting queue: %s", msg.class_name)

        self.message_q.put(msg)

    def retry_message(self, msg):
        # Never send a message more than twice (original time plus one retry)
        # To prevent queue from filling up or causing memory issues.
        if msg.sent_count < MAX_ATTEMPTS:
            logger.debug("Re-enqueuing %s message", msg.class_name)
            self.add_message(msg)


@dataclass
class StackFrame:
    """
    StackFrame represents a single stack frame used used during direct
    reporting.

    We might be able to use AssessStackFrame in the future. We're using
    this for now because it's quicker to implement.
    """

    line_number: int | None
    file_name: str
    method_name: str
    declaring_class: str | None = None
    signature: str | None = None
    type: str | None = None
    eval: str | None = None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import contrast
from contrast.agent.agent_lib.input_tracing import InputAnalysisResult
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.api.sample import Sample
from contrast.api.user_input import InputType

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class MethodTampering(BaseRule):
    RULE_NAME = "method-tampering"
    USER_INPUT_KEY = InputType.METHOD.name

    def postfilter(self):
        """
        At postfilter we generate activity if input analysis was found and depending on application response code.

        If response code is either 4xx or 5xx, application was not exploited (only probed) by an unexpected HTTP method.
        If response code is anything else, then an unexpected HTTP method successfully exploited the application.
        """
        logger.debug("PROTECT: Postfilter", rule=self.name)

        evaluations_for_rule = self.evaluations_for_rule()

        context = contrast.REQUEST_CONTEXT.get()

        # do not remove; this case is not yet well-understood
        if (
            context is None
            or not hasattr(context, "response")
            or context.response is None
        ):
            logger.debug("WARNING: failed to get context in MethodTampering.postfilter")
            return

        response_code = context.response.status_code
        if str(response_code).startswith("4") or str(response_code).startswith("5"):
            if not self.probe_analysis_enabled:
                logger.debug(
                    "PROTECT: skipping probe report",
                    reason="probe analysis disabled",
                    rule=self.name,
                )
                return
            context.attacks.extend(
                attack
                for evaluation in evaluations_for_rule
                if (
                    attack := self.build_attack_without_match(
                        evaluation=evaluation,
                        method=evaluation.input.value,
                        response_code=response_code,
                    )
                )
            )
        else:
            context.attacks.extend(
                attack
                for evaluation in evaluations_for_rule
                if (
                    attack := self.build_attack_with_match(
                        None,
                        evaluation=evaluation,
                        method=evaluation.input.value,
                        response_code=response_code,
                    )
                )
            )

    def build_sample(
        self,
        evaluation: InputAnalysisResult | None,
        candidate_string: str | None,
        **kwargs,
    ) -> Sample:
        assert evaluation is not None
        sample = self.build_base_sample(evaluation)

        sample.details["method"] = kwargs.get("method", "")
        sample.details["responseCode"] = kwargs.get("response_code", -1)

        return sample



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

ENVIRONMENT_VARIABLE_SRC = "ENVIRONMENT_VARIABLE"
USER_CONFIGURATION_FILE_SRC = "USER_CONFIGURATION_FILE"
DEFAULT_VALUE_SRC = "DEFAULT_VALUE"
CONTRAST_UI_SRC = "CONTRAST_UI"


class ConfigOption:
    def __init__(
        self,
        canonical_name,
        default_value,
        type_cast,
        redacted=False,
        log_effective_config=True,
    ):
        assert type_cast is not bool  # use str_to_bool
        self.canonical_name = canonical_name
        self.default_value = default_value
        self.type_cast = type_cast
        self.name = None
        self.override_value = None
        self.env_value = None
        self.file_values = None
        self.file_sources = None
        self.ui_value = None
        self.redacted = redacted
        self.log_effective_config = log_effective_config

    def value(self):
        if self.override_value is not None:
            return self.override_value
        if self.env_value is not None:
            return self.env_value
        if self.file_values:
            return self.file_values[0]
        if self.ui_value is not None:
            return self.ui_value
        return self.default_value

    def source(self) -> str:
        if self.env_value is not None:
            return ENVIRONMENT_VARIABLE_SRC
        if self.file_values:
            return USER_CONFIGURATION_FILE_SRC
        if self.ui_value is not None:
            return CONTRAST_UI_SRC
        return DEFAULT_VALUE_SRC

    def is_definitely_static(self) -> bool:
        """
        Static config options cannot change during their lifetime.* These options have
        been set by high-priority, unchangeable sources - ie env vars or a local config
        file.

        NOTE: The return value from this function is only meaningful when it is True. A
        return value of False indicates that the ConfigOption _might_ not be static. For
        example, an option currently using the default value will never be static
        according to this function, but if it is also not configurable via the UI, it is
        effectively static.

        *override_value potentially allows this property to be violated, but it only
        currently applies to agent.enable and should be used sparingly.
        """
        return self.override_value is not None or self.source() in {
            ENVIRONMENT_VARIABLE_SRC,
            USER_CONFIGURATION_FILE_SRC,
        }

    def file_name(self):
        return (
            self.file_sources[0]
            if self.file_sources and self.source() == USER_CONFIGURATION_FILE_SRC
            else None
        )

    def provided_name(self) -> str:
        if self.name is not None:
            return self.name
        return self.canonical_name

    def loggable_value(self) -> str:
        return self.to_string(self.value())

    def to_string(self, raw_value) -> str:
        # If the value is empty, just return empty String.
        if raw_value in (None, ""):
            return ""
        # If the option is sensitive, like an API credential, we do not log or report it.
        if self.redacted:
            return "**REDACTED**"
        str_value = str(raw_value)
        # If the option is an enum, we need to clean it up.
        if str_value and str_value.startswith("Mode."):
            return str_value.removeprefix("Mode.").upper()
        return str_value

    def clear(self):
        """
        This is used as a convenience method during testing to ensure a clean slate.
        """
        self.override_value = None
        self.env_value = None
        self.file_values = None
        self.ui_value = None
        self.file_sources = None



from django.db import models


class Message(models.Model):
    text = models.CharField(max_length=255)
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"Message(id={self.id}, text={self.text})"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Builds/defines propagation functions to be called from C extension hooks. Only
propagation functions for NECESSARY_PROPAGATORS will be defined in this module.

NOTE: Propagator functions must be generated before the C extension is initialized.
"""

import sys

from contrast.agent.assess.utils import is_tracked
from contrast_vendor import structlog as logging

import contrast
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.propagation_policy import track_copy_without_new_event

# We call these from the C extension - they're import aliases
from contrast.agent.assess.policy.propagation_policy import (  # noqa: F401
    create_stream_source_event,
    propagate_stream,
)
from contrast.agent.assess.policy.propagators import BasePropagator, PROPAGATOR_ACTIONS
from contrast.agent.policy import registry

NECESSARY_PROPAGATORS = (
    "__repr__",
    "repeat",
    "CAST",
    "subscript",
)
HOOK_TYPE_NAMES = ["unicode", "bytes", "bytearray"]

logger = logging.getLogger("contrast")
module = sys.modules[__name__]


def build_generic_strtype_propagator(policy_node):
    """
    Build propagator function used for most methods of unicode and bytes types

    By not accounting for bytearray propagation in this function we are able to
    optimize for the most frequent propagation events.

    We also optimize by avoiding handling for a few unicode/bytes special case methods
    in this function. Those special cases are handled by other builders below.

    While this leads to some code duplication, it allows us to optimize some of the
    most intensively called code in the agent.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_string(target, self_obj, ret, args, kwargs):
        # Short-circuit that applies to all string types except bytearray
        if ret is self_obj:
            return

        context = contrast.REQUEST_CONTEXT.get()
        if context is None or context.stop_propagation:
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_string


def build_bytearray_propagator(policy_node):
    """
    Build propagator function used for all methods of bytearray instances

    Copies of bytearrays return a new object, so it is necessary for us to track the
    new string, even if it hasn't changed.

    We could break the concat special case into a separate propagator function, but I'm
    not really sure it's worth it for bytearray since bytearay events are expected to
    be pretty rare. We can change it if it ever becomes a performance issue.

    Keeping this case separate from the generic builder allows us to avoid slowing down
    the propagation of the more frequently called unicode/bytes propagators, even if it
    leads to some duplicated code.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)
    method_name = policy_node.method_name

    def propagate_bytearray(target, self_obj, ret, args, kwargs):
        context = contrast.REQUEST_CONTEXT.get()
        if context is None or context.stop_propagation:
            return

        if ret == self_obj and method_name != "concat":
            track_copy_without_new_event(target, self_obj)
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_bytearray


def build_generic_cast_propagator(policy_node):
    """
    Build propagator function used for cast propagation for unicode/bytes

    Having a separate builder here allows us to optimize for the translate special case
    without imposing a blanket performance overhead on all other unicode/bytes methods.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_cast(target, self_obj, ret, args, kwargs):
        if ret is (args and args[0]) or is_tracked(ret):
            # If the return value is the same as the first argument,
            # then no new object was created and we'll already be tracking
            # the return value (or not).
            # If the return value is already tracked, we already propagated
            # to it (likely via a custom __str__ or __bytes__ method) and
            # don't need to do so again.
            return

        context = contrast.REQUEST_CONTEXT.get()
        if context is None or context.stop_propagation:
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_cast


def build_bytearray_cast_propagator(policy_node):
    """
    Build propagator function used for cast propagation for bytearray

    Having a separate builder here allows us to optimize for the translate special case
    without imposing a blanket performance overhead on all other unicode/bytes methods.
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    def propagate_cast(target, self_obj, ret, args, kwargs):
        source = args and args[0]
        if (
            ret == source
            and isinstance(target, bytearray)
            and isinstance(source, bytearray)
        ):
            track_copy_without_new_event(target, source)
            return

        context = contrast.REQUEST_CONTEXT.get()
        if context is None or context.stop_propagation:
            return

        preshift = Preshift(self_obj, args or [], kwargs or {})

        propagator = propagator_class(policy_node, preshift, target)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug(
            "Propagator %s found: propagated to %s", policy_node.name, id(target)
        )

        context.propagated()

    return propagate_cast


def _create_propagator_function(strtype, location, propagation_node, builder):
    propagator_name = f"propagate_{strtype}_{location.method_name.lower()}"
    propagator = builder(propagation_node)
    propagator.__name__ = propagator_name
    setattr(module, propagator_name, propagator)


PROPAGATION_SPECIAL_CASES = {
    ("unicode", "CAST"): build_generic_cast_propagator,
    ("bytes", "CAST"): build_generic_cast_propagator,
    ("bytearray", "CAST"): build_bytearray_cast_propagator,
}


def build_string_propagator_functions():
    """
    Builds propagator functions that will be called from C extension hooks
    """
    for location in registry.get_string_method_nodes():
        if location.method_name not in NECESSARY_PROPAGATORS:
            continue
        prop_node = location.propagator_nodes[0]
        for strtype in HOOK_TYPE_NAMES:
            # Check whether there's a special case for this hook location
            propagator = PROPAGATION_SPECIAL_CASES.get(
                (strtype, location.method_name),
                (
                    # General case for either bytearray or other strtype
                    build_bytearray_propagator
                    if strtype == "bytearray"
                    else build_generic_strtype_propagator
                ),
            )
            _create_propagator_function(strtype, location, prop_node, propagator)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class AdjustedSpan:
    def __init__(self, start, stop):
        self.start = start
        self.stop = stop

    def length(self):
        return self.stop - self.start

    def __eq__(self, other):
        if not isinstance(other, AdjustedSpan):
            return False
        return self.start == other.start and self.stop == other.stop

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.start}, {self.stop})>"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


from collections.abc import Generator, Mapping
import contextlib
from typing import Protocol


from contrast.agent.request import Request
from contrast.configuration.agent_config import AgentConfig
from contrast.reporting.teamserver_messages.base_ts_message import BaseTsMessage
import contrast_fireball


class Reporter(Protocol):
    # NOTE: server_type is separate from config because its default value
    # is generated at runtime.
    def initialize_application(self, config: AgentConfig, server_type="") -> bool: ...

    def new_discovered_routes(self, routes: set[contrast_fireball.DiscoveredRoute]): ...

    def new_observed_route(self, route: contrast_fireball.ObservedRoute): ...

    # new_findings is a batching method, but the Fireball client
    # will accept a single finding at a time and batch them internally.
    # When Fireball is the primary reporting client, we should consider
    # moving findings to a fire-and-forget model instead of batching.
    def new_findings(
        self, findings: list[contrast_fireball.AssessFinding], request: Request | None
    ): ...

    def new_libraries(self, libraries: list[contrast_fireball.Library]): ...

    def new_library_observations(
        self, observations: list[contrast_fireball.LibraryObservation]
    ): ...

    def new_inventory_components(
        self, components: list[contrast_fireball.InventoryComponent]
    ): ...

    @contextlib.contextmanager
    def observability_trace(
        self,
        *,
        send_trace: bool,
        attributes: contrast_fireball.OtelAttributes | None = None,
    ) -> Generator: ...

    # Legacy methods from direct reporting
    def add_message(self, msg: BaseTsMessage): ...

    def send_message(self, msg: BaseTsMessage): ...

    def retry_message(self, msg: BaseTsMessage): ...


def get_reporting_client(config: Mapping) -> Reporter:
    client_type = config.get("api.reporting_client")
    if client_type == "fireball":
        from contrast.reporting.fireball import Client
    elif client_type == "direct":
        from contrast.reporting.reporting_client import ReportingClient as Client
    else:
        raise ValueError(f"Invalid reporting client: {client_type}")

    return Client()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class Namespace:
    """
    Base class for creating module-level namespaces. The purpose of this is to
    circumvent private global variables, which are a hassle. This base class is
    not strictly necessary, but it's a great place for:
    - documentation
    - preventing instantiation of classes used merely as namespaces
    - making the purpose of subclasses very clear when they appear in modules

    An example use case is as follows:

    ```
    class module(Namespace):
        _module_variable_1 = {}
        _module_variable_2 = []
    ```

    Inside of the module, the variables are now available as `module._module_variable_1` etc.
    We suggest the lowercase name `module` for subclasses, as it is a natural extension of
    the `self` and `cls` conventions.

    Compare this globals; modifying a global requires the `global` keyword, which
    is both ugly and easy to forget (making it bug-prone).
    """

    def __new__(cls, *args, **kwargs):
        """
        Prevent instantiation, because that doesn't make sense here
        """
        raise TypeError("Namespace derivatives may not be instantiated")



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators import SafeJoinPropagator


class SafeJoinDjangoPropagator(SafeJoinPropagator):
    """
    Django's safe join propagation is the same as the base safe join except
    for one small detail which is the use of abspath. Django's safe_join
    function has a different behavior when relative paths are included in
    the directory.
    For example,

    ```
    >>> path = '/not/a/real/path/../path/test/'
    >>> file = 'base.html'
    >>>
    >>> flask_safe_join(path, file)
    '/not/a/real/path/../path/test/base.html'

    >>> django_safe_join(path, file)
    '/not/a/real/path/test/base.html'

    Since django simplifies the result, our propagation has to account for that
    when applying the SAFE_PATH tag to the result.
    ```
    """

    use_abspath = True



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import requests

from contrast.reporting.teamserver_messages.effective_config import EffectiveConfig

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly


class ApplicationSettings(BaseTsAppMessage):
    def __init__(self):
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"
        self.extra_headers["If-Modified-Since"] = (
            self.settings.app_settings_last_modified
        )
        self.body = None

    @property
    def name(self):
        return "application-settings"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "settings",
            ]
        )

    @property
    def expected_response_codes(self) -> list[int]:
        return [200]

    @property
    def request_method(self):
        return requests.get

    @fail_loudly("Failed to process ApplicationSettings response")
    def process_response(self, response, reporting_client):
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()
        last_modified = response.headers.get("Last-Modified", None)

        self.settings.apply_application_settings(body, last_modified)
        self.settings.log_effective_config()
        if reporting_client is not None and self.settings.is_agent_config_enabled():
            reporting_client.add_message(EffectiveConfig())



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
def assert_false():
    """
    Only used for testing
    """
    assert False  # noqa: B011 PT015



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import _io
import os

from contrast import import_hook
from contrast.agent.assess.policy import string_propagation
from contrast.agent.policy import registry
from contrast.agent.policy.applicator import register_policy_patches
from contrast.agent.policy.rewriter import apply_rewrite_policy
from contrast.agent.settings import Settings
from contrast.assess_extensions import cs_str
from contrast.patches import (
    cs_str as cs_str_patches,
)
from contrast.patches import (
    register_assess_monkeypatches,
    register_common_monkeypatches,
)
from contrast.utils.configuration_utils import str_to_bool
from contrast.utils.namespace import Namespace
from contrast.utils.patch_utils import repatch_imported_modules
from contrast_rewriter import register as register_rewriter
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class module(Namespace):
    assess_patches_enabled: bool = False


def is_preinstrument_flag_set() -> bool:
    """
    Special flag signaling that the agent should apply all instrumentation on startup
    regardless of which modes are enabled. This allows any mode to be enabled later
    without requiring an application restart.
    """
    try:
        return str_to_bool(os.environ.get("CONTRAST_PREINSTRUMENT"))
    except ValueError:
        return False


def _enable_pure_python_string_patches(string_types):
    """
    Enable string propagation hooks for individual string type methods

    This method uses policy to determine all of the string methods that need to be
    patched. Patches are applied in pure Python.
    """
    for strtype in string_types:
        for node in registry.get_string_method_nodes():
            method_name = node.method_name

            if method_name.lower() in ["cast", "concat"]:
                # these are applied directly in the C extension only
                continue

            real_method_name = (
                "format_map" if method_name == "formatmap" else method_name
            )

            if not hasattr(strtype, real_method_name):
                continue

            cs_str_patches.patch_strtype_method(strtype, real_method_name)
            logger.debug(
                "Applied pure Python patch for %s.%s", strtype.__name__, method_name
            )


def _enable_pure_python_stream_patches():
    cs_str_patches.patch_stream_method(_io._IOBase, "readline")
    cs_str_patches.patch_stream_method(_io._IOBase, "readlines")
    cs_str_patches.patch_stream_method(_io.BytesIO, "read")
    cs_str_patches.patch_stream_method(_io.BytesIO, "readline")
    cs_str_patches.patch_stream_method(_io.BytesIO, "readlines")
    cs_str_patches.patch_stream_method(_io.BytesIO, "read1")
    cs_str_patches.patch_stream_method(_io.StringIO, "read")
    cs_str_patches.patch_stream_method(_io.StringIO, "readline")


def enable_assess_patches():
    """
    Enables string patches.

    Has no effect if these patches are already enabled.
    """
    if module.assess_patches_enabled:
        return
    # NOTE: String propagator functions *must* be built before the extension
    # is initialized.
    string_propagation.build_string_propagator_functions()
    cs_str.initialize()
    cs_str_patches.enable_str_properties()
    cs_str.enable_c_patches()
    _enable_pure_python_string_patches([str, bytes, bytearray])
    _enable_pure_python_stream_patches()

    module.assess_patches_enabled = True


def disable_assess_patches():
    """
    Disables extension hooks and other string patches.

    Has no effect if these patches are not already enabled.

    This does not disable "pure python" strtype patches applied with set_attr_on_type.
    """
    if not module.assess_patches_enabled:
        return

    cs_str.disable_c_patches()

    module.assess_patches_enabled = False


def _enable_protect_patches():
    register_common_monkeypatches()

    logger.debug("adding protect policy")
    register_policy_patches(protect_mode=True)

    # This has no effect if the patches are not enabled
    disable_assess_patches()


def _enable_assess_patches():
    enable_assess_patches()

    # Policy-based rewrites need to be applied prior to any policy patches.
    # Policy patches can be layered on top of rewritten functions. So that
    # means we need to make sure that the "original" function called by the
    # policy patch is the *rewritten* one.
    # Pathlib rewrites must not be applied here. They are only stable when applied by
    # the runner. The exact reason for this is unknown, but it's related to repatching.
    # If this causes noticeable issues, use the runner. This will eventually be
    # mandatory anyway, and policy-based rewrites will be removed from here entirely.
    apply_rewrite_policy(rewrite_pathlib=False)

    logger.debug("enabled assess string patches")
    register_common_monkeypatches()
    register_assess_monkeypatches()

    logger.debug("adding assess policy")
    register_policy_patches(protect_mode=False)

    # This is included as a fallback so that we continue to support the case
    # where the runner is not used. If the rewriter has already been enabled by
    # the runner, this has no effect (other than to log a message).
    register_rewriter()


def enable_patches(*, preinstrument: bool):
    settings = Settings()

    if settings.is_analyze_libs_enabled() or preinstrument:
        import_hook.register_path_finder()

    from contrast.agent import agent_state

    if agent_state.module.protect_enabled or preinstrument:
        _enable_protect_patches()
    if agent_state.module.assess_enabled or preinstrument:
        _enable_assess_patches()
        import_hook.register_path_finder()
    if agent_state.module.observe_enabled or preinstrument:
        register_policy_patches(protect_mode=False)
        # register import hook to deadzone file-open-create action
        # reports for source files.
        import_hook.register_path_finder()

    logger.debug("revisiting imported modules to apply patches")
    repatch_imported_modules()
